#!/usr/bin/env python3
# Copyright (C) 2025 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Test dtype conversions."""

import numpy as np
import pytest

from polar2grid.core.dtype import dtype_to_str, str_to_dtype


@pytest.mark.parametrize(
    ("input_arg", "exp_str"),
    [
        (np.uint8, "uint8"),
        ("uint8", "uint8"),
        ("u1", "uint8"),
        (np.dtype("uint8"), "uint8"),
        (np.float32, "float32"),
        ("float32", "float32"),
        ("f4", "float32"),
        (np.dtype("float32"), "float32"),
        (np.float64, "float64"),
        ("float64", "float64"),
        ("f8", "float64"),
        (np.dtype("float64"), "float64"),
        (np.int8, "int8"),
        ("int8", "int8"),
        ("i1", "int8"),
        (np.dtype("int8"), "int8"),
    ],
)
def test_dtype_to_str(input_arg, exp_str):
    """Test conversion from dtype to string."""
    dtype_str = dtype_to_str(input_arg)
    assert dtype_str == exp_str


@pytest.mark.parametrize(
    ("input_arg", "exp_dtype"),
    [
        ("uint8", np.uint8),
        ("int8", np.int8),
        ("float32", np.float32),
    ],
)
def test_str_to_dtype(input_arg, exp_dtype):
    """Test conversion from string to dtype."""
    np_dtype = str_to_dtype(input_arg)
    assert np_dtype == exp_dtype



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2012-2015 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
# Written by David Hoese    December 2014
# University of Wisconsin-Madison
# Space Science and Engineering Center
# 1225 West Dayton Street
# Madison, WI  53706
# david.hoese@ssec.wisc.edu

"""The HDF5 writer creates HDF5 files with groups for each gridded area.

All selected products are in one file.
Products are subgrouped together under a parent HDF5 data group
based on the data product projection/remapping (parent projection group).
Each parent projection group contains attributes describing the projection.
Product subgroups contain attributes of the data including timestamps,
sensor and platform information.
See the command line arguments for HDF5 compression options, the flag to include
longitude and latitude data in the file, instructions for output-filename
patterns, and product selection.
"""

from __future__ import annotations

import logging
import os
from typing import TextIO

import h5py
import numpy as np
import xarray as xr
from pyresample.geometry import SwathDefinition
from satpy.writers.core.base import Writer
from satpy.writers.core.compute import compute_writer_results, split_results

from polar2grid.utils.legacy_compat import convert_p2g_pattern_to_satpy
from polar2grid.utils.warnings import ignore_pyproj_proj_warnings
from polar2grid.writers.geotiff import NUMPY_DTYPE_STRS, NumpyDtypeList, str_to_dtype

LOG = logging.getLogger(__name__)

# reader_name -> filename
DEFAULT_OUTPUT_FILENAMES = {
    "polar2grid": {
        None: "{platform_name}_{sensor}_{start_time:%Y%m%d_%H%M%S}.h5",
    },
    "geo2grid": {
        None: "{platform_name}_{sensor}_{start_time:%Y%m%d_%H%M%S}.h5",
    },
}


def all_equal(iterable: list[str]) -> bool:
    """Return True if all the elements are equal to each other."""
    return all(iterable[0] == item for item in iterable[1:])


class FakeHDF5:
    """Use fake hdf class to create targets for da.store and delayed sources."""

    def __init__(self, output_filename: str, var_name: str):
        """Initialize filename target with appropriate var_name for the data."""
        self.output_filename = output_filename
        self.var_name = var_name

    def __setitem__(self, write_slice, data):
        """Write data arrays to HDF5 file either delayed or not."""
        with h5py.File(self.output_filename, mode="a") as fh:
            fh[self.var_name][write_slice] = data


class HDF5Writer(Writer):
    """Writer for HDF5 files."""

    def __init__(self, **kwargs):
        """Init the writer."""
        super(HDF5Writer, self).__init__(**kwargs)

        if self.filename_parser is None:
            raise RuntimeError("No filename pattern or specific filename provided")

    def _output_file_kwargs(self, dataset, dtype):
        """Get file keywords from data for output_pattern."""
        if isinstance(dataset, list):
            dataset = dataset[0]
            area = dataset[0].attrs["area"]
            d_dtype = dataset[0].dtype
        else:
            area = dataset.attrs["area"]
            d_dtype = dataset.dtype

        dtype = d_dtype if dtype is None else dtype

        args = dataset.attrs
        args["grid_name"] = "native" if isinstance(area, SwathDefinition) else area.area_id
        args["rows"], args["columns"] = area.shape
        args["data_type"] = dtype

        return args

    def iter_by_area(self, datasets: list[xr.DataArray]):
        """Generate datasets grouped by Area.

        Args:
            datasets (list[xr.DataArray]):  A list of dataArray objects stored in Scene.

        Returns:
            dictionary:  a dictionary of {AreaDef:  list[xr.DataArray]}
        """
        datasets_by_area = {}
        for ds in datasets:
            a = ds.attrs.get("area")
            datasets_by_area.setdefault(a, []).append(ds)
        return datasets_by_area.items()

    @staticmethod
    def open_HDF5_filehandle(output_filename: str, append: bool = True):
        """Open a HDF5 file handle."""
        if os.path.isfile(output_filename):
            if append:
                LOG.info("Appending to existing file: %s", output_filename)
                mode = "a"
            else:
                LOG.warning("HDF5 file already exists, will overwrite/truncate: %s", output_filename)
                mode = "w"
        else:
            LOG.info("Creating HDF5 file: %s", output_filename)
            mode = "w"

        h5_fh = h5py.File(output_filename, mode)
        return h5_fh

    @staticmethod
    def create_proj_group(filename: str, parent: TextIO, area_def):
        """Create the top level group from projection information."""
        projection_name = (area_def.area_id).replace(" ", "_")
        # if top group alrady made, return.
        if projection_name in parent:
            return projection_name

        # create top group for first time
        group = parent.create_group(projection_name)
        # add attributes from grid_defintion.
        if isinstance(area_def, SwathDefinition):
            group.attrs["height"], group.attrs["width"] = area_def.shape
            group.attrs["description"] = "No projection: native format"
        else:
            with ignore_pyproj_proj_warnings():
                group.attrs["proj4_definition"] = area_def.crs.to_string()
            for a in ["height", "width"]:
                ds_attr = getattr(area_def, a, None)
                if ds_attr is None:
                    pass
                else:
                    group.attrs[a] = ds_attr

            group.attrs["cell_height"] = np.round(-area_def.pixel_size_y, 5)
            group.attrs["cell_width"] = np.round(area_def.pixel_size_x, 5)
            group.attrs["origin_x"] = area_def.pixel_upper_left[0]
            group.attrs["origin_y"] = area_def.pixel_upper_left[1]

        return projection_name

    def write_geolocation(
        self, fh, fname: str, parent: str, area_def, dtype: np.dtype, append: bool, compression, chunks: tuple[int, int]
    ) -> tuple[list, list[FakeHDF5]]:
        """Delayed Geolocation Data write."""
        msg = ("Adding geolocation 'longitude' and 'latitude' datasets for grid %s", parent)
        LOG.info(msg)
        lon_data, lat_data = area_def.get_lonlats(chunks=chunks)

        dtype = lon_data.dtype if dtype is None else dtype
        data_shape = lon_data.shape

        lon_grp = "{}/longitude".format(parent)
        lat_grp = "{}/latitude".format(parent)

        if append:
            for var_name in [lon_grp, lat_grp]:
                if var_name in fh:
                    LOG.warning("Product %s already exists in HDF5 group, will delete existing dataset", var_name)
                    del fh[var_name]

        lon_dataset = FakeHDF5(fname, lon_grp)
        lat_dataset = FakeHDF5(fname, lat_grp)
        fh.create_dataset(lon_grp, shape=data_shape, dtype=dtype, compression=compression)
        fh.create_dataset(lat_grp, shape=data_shape, dtype=dtype, compression=compression)

        return [lon_data, lat_data], [lon_dataset, lat_dataset]

    @staticmethod
    def create_variable(hdf_fh, hdf_subgroup: str, data_arr: xr.DataArray, dtype: np.dtype, compression: bool):
        """Create a HDF5 data variable and attributes for the variable."""
        ds_attrs = data_arr.attrs

        d_dtype = data_arr.dtype if dtype is None else dtype

        if hdf_subgroup in hdf_fh:
            LOG.warning("Product %s already in HDF5 group,will delete existing dataset", hdf_subgroup)
            del hdf_fh[hdf_subgroup]

        dset = hdf_fh.create_dataset(hdf_subgroup, shape=data_arr.shape, dtype=d_dtype, compression=compression)

        dset.attrs["satellite"] = ds_attrs["platform_name"]
        dset.attrs["instrument"] = ds_attrs["sensor"]
        dset.attrs["begin_time"] = ds_attrs["start_time"].isoformat()
        dset.attrs["end_time"] = ds_attrs["end_time"].isoformat()

    def save_datasets(
        self,
        dataset: list[xr.DataArray],
        filename=None,
        dtype=None,
        append=True,
        compute=True,
        **kwargs,
    ):
        """Save HDF5 datasets."""
        compression = kwargs.pop("compression", None)
        if compression == "none":
            compression = None

        add_geolocation = kwargs.pop("add_geolocation", False)

        # will this be written to one or multiple files?
        output_names = []

        for dataset_id in dataset:
            file_attrs = self._output_file_kwargs(dataset_id, dtype)
            out_filename = filename or self.get_filename(**file_attrs)
            output_names.append(out_filename)

        filename = output_names[0]
        if not all_equal(output_names):
            LOG.warning("More than one output filename possible. Writing to only '{}'.".format(filename))

        HDF5_fh = self.open_HDF5_filehandle(filename, append=append)

        datasets_by_area = self.iter_by_area(dataset)
        # Initialize source/targets at start of each new AREA grouping.
        dsets = []
        targets = []

        for area, data_arrs in datasets_by_area:
            dask_arrays, file_targets = self._save_data_arrays_and_area(
                area,
                data_arrs,
                filename,
                HDF5_fh,
                dtype,
                append,
                compression,
                add_geolocation,
            )
            dsets.extend(dask_arrays)
            targets.extend(file_targets)

        results = (dsets, targets)
        if compute:
            LOG.info("Computing and writing results...")
            return compute_writer_results([results])

        targets, sources, delayeds = split_results([results])
        if delayeds:
            # This writer had only delayed writes
            return delayeds
        else:
            return targets, sources

    def _save_data_arrays_and_area(
        self, area, data_arrs, filename, HDF5_fh, dtype, append, compression, add_geolocation
    ):
        # open HDF5 file handle, check if group already exists.
        parent_group = self.create_proj_group(filename, HDF5_fh, area)

        dsets = []
        targets = []
        if add_geolocation:
            chunks = data_arrs[0].chunks
            geo_sets, file_targets = self.write_geolocation(
                HDF5_fh, filename, parent_group, area, dtype, append, compression, chunks
            )
            dsets.extend(geo_sets)
            targets.extend(file_targets)

        for data_arr in data_arrs:
            try:
                dask_arr, target_file = self._save_data_array(
                    HDF5_fh, filename, data_arr, parent_group, dtype, compression
                )
            except ValueError:
                if os.path.isfile(filename):
                    os.remove(filename)
                raise
            dsets.append(dask_arr)
            targets.append(target_file)
        return dsets, targets

    def _save_data_array(self, HDF5_fh, filename, data_arr, parent_group, dtype, compression):
        hdf_subgroup = "{}/{}".format(parent_group, data_arr.attrs.get("p2g_name", data_arr.attrs["name"]))
        file_var = FakeHDF5(filename, hdf_subgroup)
        self.create_variable(HDF5_fh, hdf_subgroup, data_arr, dtype, compression)
        return data_arr.data, file_var


def add_writer_argument_groups(parser, group=None):
    """Create writer argument groups."""
    if group is None:
        group = parser.add_argument_group(title="HDF5 Writer")
    group.add_argument(
        "--output-filename",
        dest="filename",
        type=convert_p2g_pattern_to_satpy,
        help="Custom file pattern to save dataset to",
    )
    group.add_argument(
        "--dtype",
        choices=NumpyDtypeList(NUMPY_DTYPE_STRS),
        type=str_to_dtype,
        help="Data type of the output file (8-bit unsigned integer by default - uint8)",
    )
    group.add_argument(
        "--compress",
        dest="compression",
        choices=["none", "gzip", "lzf"],  # , "szip"],
        default="none",
        help="Dataset compression algorithm. Defaults to no compression.",
    )
    group.add_argument(
        "--add-geolocation",
        action="store_true",
        help="Add 'longitude' and 'latitude' datasets for each grid",
    )
    group.add_argument(
        "--no-append",
        dest="append",
        action="store_false",
        help="Don't append to the HDF5 file if it already exists (otherwise may overwrite data)",
    )

    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Tests for the binary writer."""

import os
from datetime import datetime
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import satpy
import xarray as xr
from satpy import Scene

TEST_ETC_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "etc"))


def _int_min_max(dtype):
    info = np.iinfo(dtype)
    return info.min, info.max


def _min_max_for_two_dtypes(src_dtype, dst_dtype):
    src_is_int = not np.issubdtype(src_dtype, np.floating)
    dst_is_int = not np.issubdtype(dst_dtype, np.floating)
    if src_is_int and dst_is_int:
        smin, smax = _int_min_max(src_dtype)
        dmin, dmax = _int_min_max(dst_dtype)
        return max(smin, dmin), min(smax, dmax)
    elif src_is_int:
        return _int_min_max(src_dtype)
    return _int_min_max(dst_dtype)


def _create_fake_data_arr(shape=(100, 50), dims=("y", "x"), dtype=np.float64):
    area_mock = mock.MagicMock()
    area_mock.area_id = "fake_area"
    attrs = {
        "name": "fake_name",
        "p2g_name": "fake_p2g_name",
        "platform_name": "NOAA-20",
        "sensor": "viirs",
        "start_time": datetime(2021, 1, 1, 12, 0, 0),
        "end_time": datetime(2021, 1, 1, 12, 10, 0),
        "area": area_mock,
        "standard_name": "test_arange",
    }
    data = np.arange(np.prod(shape), dtype=np.float64)
    if not np.issubdtype(dtype, np.floating):
        data = np.clip(data, *_int_min_max(dtype))
        attrs["_FillValue"] = 0
    data = data.astype(dtype)
    data = da.from_array(data, chunks=10).reshape(shape)
    data_arr = xr.DataArray(data, attrs=attrs, dims=dims)
    return data_arr


class TestBinaryWriter:
    """Test the binary writer."""

    def setup_method(self):
        """Add P2G configs to the Satpy path."""
        from polar2grid.utils.config import add_polar2grid_config_paths

        self._old_path = satpy.config.get("config_path")
        add_polar2grid_config_paths()
        # add test specific configs
        curr_path = satpy.config.get("config_path")
        satpy.config.set(config_path=[TEST_ETC_DIR] + curr_path)

    def teardown_method(self):
        """Reset Satpy config path back to the original value."""
        satpy.config.set(config_path=self._old_path)

    @pytest.mark.parametrize("enhance", [True, False])
    @pytest.mark.parametrize("dst_dtype", [None, np.float32, np.float64, np.uint8])
    @pytest.mark.parametrize("src_dtype", [np.float32, np.float64, np.uint8])
    def test_basic_write(self, tmpdir, src_dtype, dst_dtype, enhance):
        """Test writing data to disk."""
        src_data_arr = _create_fake_data_arr(dtype=src_dtype)
        scn = Scene()
        scn[src_data_arr.attrs["name"]] = src_data_arr
        scn.save_datasets(writer="binary", base_dir=str(tmpdir), dtype=dst_dtype, enhance=enhance)
        exp_fn = tmpdir.join("noaa-20_viirs_fake_p2g_name_20210101_120000_fake_area.dat")

        if dst_dtype is None:
            dst_dtype = src_dtype if src_dtype != np.float64 else np.float32
        assert os.path.isfile(exp_fn)
        data = np.memmap(str(exp_fn), mode="r", dtype=dst_dtype)
        exp_data = self._generate_expected_output(src_data_arr, dst_dtype, enhance)
        np.testing.assert_allclose(data, exp_data, atol=2e-7)

    @staticmethod
    def _generate_expected_output(src_data_arr, dst_dtype, enhance):
        src_dtype = src_data_arr.dtype
        exp_data = src_data_arr.values.astype(np.float64).ravel()
        if enhance:
            exp_data = exp_data / 500.0  # see polar2grid/tests/etc/enhancements/generic.yaml
            if not np.issubdtype(dst_dtype, np.floating):
                rmin, rmax = _int_min_max(dst_dtype)
                exp_data = exp_data * (rmax - rmin) + rmin
                exp_data = np.clip(exp_data, rmin, rmax)
            exp_data = exp_data.astype(dst_dtype)
        elif not np.issubdtype(dst_dtype, np.floating) or not np.issubdtype(src_dtype, np.floating):
            rmin, rmax = _min_max_for_two_dtypes(src_dtype, dst_dtype)
            exp_data = np.clip(exp_data, rmin, rmax)
        return exp_data



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The ABI Level 2 Reader operates on NOAA Level 2 (L2) NetCDF files
from the GOES-16 (GOES-East), GOES-17/18 (GOES-West) Advanced Baseline
Imager (ABI) instrument. The ABI L2 NetCDF reader works off of the input
filenames to determine if a file is supported by Geo2Grid. Files usually have
the following naming scheme::

    OR_ABI-L2-{PROD}F-M3_G{SAT}_s20182531700311_e20182531711090_c20182531711149.nc

and::

    CG_ABI-L2-{PROD}F-M6_G{SAT}_s20223271830316_e20223271839394_c20223271842100.nc

These are the mission compliant file naming conventions
used by the NOAA Comprehensive Large Array-data Stewardship
System (CLASS) archive and the CSPP Geo AIT Framework Level 2 software.
The ABI L2 NetCDF reader supports most L2 products, but Geo2Grid is tested with
a limited subset of these. See the table below for more information.
The ABI L2 NetCDF reader can be provided to the main geo2grid.sh script
using the ``-r`` option and the reader name ``abi_l2_nc``.


+---------------------------+-----------------------------------------------------+
| **Product Name**          | **Description**                                     |
+===========================+=====================================================+
| AOD                       | Aerosol Optical Depth                               |
+---------------------------+-----------------------------------------------------+
| LST                       | Land Surface Temperature                            |
+---------------------------+-----------------------------------------------------+
| HT                        | Cloud Top Height                                    |
+---------------------------+-----------------------------------------------------+
| TEMP                      | Cloud Top Temperature                               |
+---------------------------+-----------------------------------------------------+
| Fog_Depth                 | Fog Depth                                           |
+---------------------------+-----------------------------------------------------+
| IFR_Fog_Prob              | Instrument Flight Rules Probability                 |
+---------------------------+-----------------------------------------------------+
| LIFR_Fog_Prob             | Low Instrument Flight Rules Probability             |
+---------------------------+-----------------------------------------------------+
| MVFR_Fog_Prob             | Marginal Visible Flight Rules Probability           |
+---------------------------+-----------------------------------------------------+

More information on the flight rules products can be found at:

https://www.experimentalaircraft.info/wx/colors-metar-taf.php

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from ._base import ReaderProxyBase
from ..core.script_utils import BooleanFilterAction

PREFERRED_CHUNK_SIZE: int = 1356

READER_PRODUCTS = [
    "AOD",
    "HT",
    "LST",
    "TEMP",
    "Fog_Depth",
    "IFR_Fog_Prob",
    "LIFR_Fog_Prob",
    "MVFR_Fog_Prob",
    "PRES",
    "Phase",
]
COMPOSITE_PRODUCTS = []


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_geo2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="ABI L2 Reader")
    group.add_argument(
        "--filter-sst",
        action=BooleanFilterAction,
        dest="filters",
        const="good_quality_qf",
        default=["good_quality_qf"],
        help="Enable or disable quality flag filtering of the SST product (default on)",
    )
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Utilities for handling legacy behavior with Satpy interfaces."""

from __future__ import annotations

import logging
from typing import Generator, Iterable, Optional, Union

from satpy import DataID, DataQuery, Scene, DatasetDict

logger = logging.getLogger(__name__)


def convert_p2g_pattern_to_satpy(pattern):
    """Convert old P2G output patterns to new format."""
    replacements = {
        "satellite": "platform_name",
        "instrument": "sensor",
        "product_name": "p2g_name",
        "grid_name": "area.area_id",
    }
    replacements.update(_generate_old_p2g_date_formats())

    for p2g_kw, satpy_kw in replacements.items():
        pattern = _warn_on_format_replaced(pattern, p2g_kw, satpy_kw)
    return pattern


def _generate_old_p2g_date_formats():
    """Convert old P2G output pattern date formats."""
    dt_frmts = {
        "_YYYYMMDD": ":%Y%m%d",
        "_YYMMDD": ":%y%m%d",
        "_HHMMSS": ":%H%M%S",
        "_HHMM": ":%H%M",
    }
    date_replacements = {
        "begin_time": "start_time:%Y%m%d_%H%M%S",
        "end_time": "end_time:%Y%m%d_%H%M%S",
    }

    for old_frmt, new_frmt in dt_frmts.items():
        old_start = "begin_time{}".format(old_frmt)
        new_start = "start_time{}".format(new_frmt)
        date_replacements[old_start] = new_start

        old_end = "end_time{}".format(old_frmt)
        new_end = "end_time{}".format(new_frmt)
        date_replacements[old_end] = new_end
    return date_replacements


def _warn_on_format_replaced(pattern: str, old_format_param: str, new_format_param: str) -> str:
    old_format_part = f"{{{old_format_param}}}"
    if old_format_part in pattern:
        logger.warning(
            f"Filename pattern includes legacy parameter: {old_format_param!r}. "
            f"Use {new_format_param!r} to avoid this warning."
        )
        return pattern.replace(old_format_part, f"{{{new_format_param}}}")
    return pattern


class AliasHandler:
    """Utility class for converting internal Satpy names to user-facing Polar2Grid names.

    Polar2Grid and Geo2Grid currently only allows users to specify the name of
    products to load as a string. For compatibility with older versions of
    Polar2Grid or for basic aesthetic differences, Polar2Grid sometimes has
    different names for products. There are also cases where Polar2Grid needs
    to provide a single name for a more complex Satpy product that normally
    requires a `DataID`.

    This class provides the ability to convert between these different forms
    of naming. If a user provides a Satpy name then this class will **not**
    replace it with a P2G alias (if it exists). This should lead to the least
    amount of surprises for users.

    """

    def __init__(
        self,
        all_aliases: dict[str, Union[str, DataQuery]],
        user_products: list[str],
    ):
        self._all_aliases = all_aliases
        self._user_products = self._unique_ordered_list(user_products)

    @staticmethod
    def _unique_ordered_list(orig_list):
        seen = set()
        return [item for item in orig_list if not (item in seen or seen.add(item))]

    def remove_unknown_user_products(
        self,
        known_dataset_names: Iterable[str],
    ) -> list[str]:
        """Remove user products that aren't known to the Scene.

        This is for a Satpy reader that dynamically determines what datasets
        it knows about based on what is in the files it is provided. If the
        user (or more likely a P2G reader's default products) asks for a
        product that isn't known, we need to remove it so Satpy doesn't error
        out. For example, the MiRS reader doesn't know what BT bands are known
        until it knows what files are being read. Polar2Grid needs to default
        to all possible BT bands or else we can't load specific BT bands.

        """
        satpy_names = self.convert_p2g_name_to_satpy(self._user_products)
        new_user_products = []
        for user_name, satpy_name in zip(self._user_products, satpy_names, strict=True):
            # convert DataID/DataQuery to string
            satpy_name = satpy_name if isinstance(satpy_name, str) else satpy_name["name"]
            if satpy_name not in known_dataset_names:
                continue
            new_user_products.append(user_name)
        self._user_products = new_user_products
        return new_user_products

    @property
    def satpy_ids(self):
        """All user provided product names as Satpy identifiers."""
        return list(self.convert_p2g_name_to_satpy())

    def convert_p2g_name_to_satpy(
        self,
        products: Optional[Iterable[str]] = None,
    ) -> Generator[Union[str, DataID], None, None]:
        """Convert P2G names to corresponding Satpy name or DataID."""
        if products is None:
            products = self._user_products
        for product_name in products:
            yield self._all_aliases.get(product_name, product_name)

    def convert_satpy_to_p2g_name(
        self,
        satpy_products: Iterable[Union[DataID]],
        possible_p2g_names: Optional[list[str]] = None,
    ):
        """Get the P2G name for a series of Satpy names or DataIDs.

        If a Satpy DataID does not have a Polar2Grid compatible name then
        ``None`` is yielded. A name is not compatible if requesting it would
        produce a different product than the original DataID.

        """
        from satpy import DatasetDict

        if possible_p2g_names is None:
            possible_p2g_names = self._user_products
        satpy_id_dict = DatasetDict({x: x for x in satpy_products})
        satpy_id_to_p2g_name = {}
        for p2g_name in possible_p2g_names:
            satpy_data_query = self._all_aliases.get(p2g_name, p2g_name)
            matching_satpy_id = _get_matching_satpy_id(satpy_id_dict, satpy_data_query)
            if matching_satpy_id is None:
                # no match
                continue

            if matching_satpy_id in satpy_id_to_p2g_name:
                logger.warning("Multiple product names map to the same identifier in Satpy")
            satpy_id_to_p2g_name[matching_satpy_id] = p2g_name

        for satpy_product in satpy_products:
            satpy_id_name = satpy_product["name"]
            satpy_id_as_p2g_name = satpy_id_to_p2g_name.get(satpy_product)
            satpy_name_is_p2g_name = satpy_id_name in possible_p2g_names
            satpy_name_does_not_round_trip = satpy_id_dict[satpy_product] != satpy_product
            if satpy_id_as_p2g_name is None:
                # We can't yield this name if it is also a P2G name or if
                # asking Satpy for the name doesn't return the same DataID
                # product. Otherwise users would ask for X and not get X.
                if satpy_name_is_p2g_name or satpy_name_does_not_round_trip:
                    yield None
                else:
                    yield satpy_id_name
                continue

            yield satpy_id_to_p2g_name.get(satpy_product, satpy_id_name)

    def apply_p2g_name_to_scene(
        self,
        scn: Scene,
    ):
        """Assign a new 'p2g_name' string attribute to every DataArray.

        This is typically done just before writing the 'Scene' output to disk so
        that the `filename` pattern used in `Scene.save_datasets` has access to
        the user-facing "P2G name" rather than the internal Satpy name or DataID.

        """
        all_ids = list(scn.keys())
        all_p2g_names = list(self.convert_satpy_to_p2g_name(all_ids))
        for data_id, p2g_name in zip(all_ids, all_p2g_names, strict=True):
            if p2g_name is None:
                # the Satpy ID doesn't have a Polar2Grid compatible name
                logger.debug("Satpy DataID %s does not have a compatible polar2grid name.", data_id)
                continue
            scn[data_id].attrs["p2g_name"] = p2g_name
            logger.debug("Mapping Satpy ID to P2G name: %s -> %s", data_id, p2g_name)

    def available_product_names(
        self, all_p2g_products: list[str], available_custom_ids: list[DataID], available_satpy_ids: list[DataID]
    ) -> tuple[list[str], list[str], list[str]]:
        """Get separate lists of available Satpy products and Polar2Grid products."""
        available_ids_as_p2g_names = list(self.convert_satpy_to_p2g_name(available_satpy_ids, all_p2g_products))
        satpy_id_to_p2g_name = dict(zip(available_satpy_ids, available_ids_as_p2g_names, strict=True))
        available_p2g_names = []
        available_custom_names = []
        available_satpy_names = []
        for satpy_id in available_satpy_ids:
            p2g_name = satpy_id_to_p2g_name[satpy_id]
            if p2g_name is None:
                # no Polar2Grid compatible name
                continue
            if p2g_name in all_p2g_products:
                available_p2g_names.append(p2g_name)
            elif satpy_id in available_custom_ids:
                available_custom_names.append(p2g_name)
            else:
                available_satpy_names.append(satpy_id["name"])

        available_satpy_names = sorted(set(available_satpy_names))
        return available_p2g_names, available_custom_names, available_satpy_names


def _get_matching_satpy_id(satpy_id_dict: DatasetDict, satpy_data_query: DataQuery | str) -> DataID:
    matching_satpy_id = None
    while matching_satpy_id is None:
        # iterate until no modifiers exist in the query and we still can't find it
        try:
            matching_satpy_id = satpy_id_dict[satpy_data_query]
            return matching_satpy_id
        except KeyError:
            if isinstance(satpy_data_query, str) or not satpy_data_query.is_modified():
                break
            satpy_data_query = satpy_data_query.create_less_modified_query()
    return matching_satpy_id


_SENSOR_ALIASES = {
    "avhrr-3": "avhrr",
}


def get_sensor_alias(satpy_sensor):
    if not isinstance(satpy_sensor, set):
        satpy_sensor = {satpy_sensor}
    new_sensor = {_SENSOR_ALIASES.get(sname, sname) for sname in satpy_sensor}
    if len(new_sensor) == 1:
        return new_sensor.pop()
    return new_sensor



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Basic usability tests for the add_coastlines script."""

from __future__ import annotations

import contextlib
import os
from unittest import mock

import numpy as np
import pytest
import rasterio
from PIL import Image

REDS_SPREAD_CMAP = {
    0: (0, 0, 0, 255),
    128: (128, 0, 0, 255),
    255: (255, 0, 0, 255),
}
REDS_MIN_CMAP = {
    0: (0, 0, 0, 255),
    1: (128, 0, 0, 255),
    2: (255, 0, 0, 255),
}


def test_add_coastlines_help():
    from polar2grid.add_coastlines import main

    with pytest.raises(SystemExit) as e:
        main(["--help"])
    assert e.value.code == 0


def _shared_fake_geotiff_kwargs(num_bands):
    kwargs = {
        "driver": "GTiff",
        "height": 1000,
        "width": 500,
        "count": num_bands,
        "dtype": np.uint8,
        "crs": "+proj=latlong",
        "transform": (0.033, 0.0, 0.0, 0.0, 0.033, 0.0),
    }
    return kwargs


def _shared_fake_l_geotiff_data(colormap: dict[int, tuple]):
    colormap_values = sorted(colormap.keys()) if colormap is not None else [0, 128, 255]
    data = np.zeros((500, 1000), dtype=np.uint8)
    data[:] = colormap_values[0]
    data[200:300, :] = colormap_values[1]
    data[300:, :] = colormap_values[2]
    return data


def _shared_fake_rgb_geotiff_data(colormap: dict[int, tuple]):
    colormap_values = [color[0] for color in colormap.values()]
    data = np.zeros((500, 1000), dtype=np.uint8)
    data[:] = colormap_values[0]
    data[200:300, :] = colormap_values[1]
    data[300:, :] = colormap_values[2]

    r_data = data
    g_data = np.zeros_like(r_data)
    b_data = np.zeros_like(r_data)
    return r_data, g_data, b_data


def _create_fake_l_geotiff(fp, colormap=None, include_scale_offset=False, include_colormap_tag=False):
    kwargs = _shared_fake_geotiff_kwargs(1)
    with rasterio.open(fp, "w", **kwargs) as ds:
        ds.write(_shared_fake_l_geotiff_data(colormap), 1)
        if include_scale_offset:
            ds.update_tags(scale=0.5, offset=0.0)
        # include_colormap_tag is not used because there is no logical way to
        # create a L geotiff that also has no colormap


def _create_fake_rgb_geotiff(fp, colormap=None, include_scale_offset=False, include_colormap_tag=False):
    kwargs = _shared_fake_geotiff_kwargs(3)
    with rasterio.open(fp, "w", **kwargs) as ds:
        r_data, g_data, b_data = _shared_fake_rgb_geotiff_data(colormap)
        ds.write(r_data, 1)
        ds.write(g_data, 2)
        ds.write(b_data, 3)
        if include_scale_offset:
            ds.update_tags(scale=0.5, offset=0.0)
        if include_colormap_tag:
            ds.update_tags(**_create_csv_cmap_and_extra_tags(colormap))


def _create_fake_l_geotiff_colormap(fp, colormap=None, include_scale_offset=False, include_colormap_tag=False):
    kwargs = _shared_fake_geotiff_kwargs(1)
    with rasterio.open(fp, "w", **kwargs) as ds:
        ds.write(_shared_fake_l_geotiff_data(colormap), 1)
        if colormap is not None:
            ds.write_colormap(1, colormap)
        if include_scale_offset:
            ds.update_tags(scale=0.5, offset=0.0)
        if colormap is not None and include_colormap_tag:
            ds.update_tags(**_create_csv_cmap_and_extra_tags(colormap))


def _create_csv_cmap_and_extra_tags(colormap):
    cmap_csv = [",".join(str(x) for x in [v] + list(color)) for v, color in colormap.items()]
    cmap_csv_str = "\n".join(cmap_csv)
    return {"colormap": cmap_csv_str}


@pytest.mark.parametrize(
    ("gen_func", "include_scale_offset", "include_cmap_tag", "output_ext"),
    [
        (_create_fake_l_geotiff, False, False, ""),
        (_create_fake_l_geotiff, False, False, "png"),
        (_create_fake_l_geotiff, False, False, "tif"),
        (_create_fake_l_geotiff_colormap, False, False, ""),
        (_create_fake_l_geotiff_colormap, True, False, ""),
        (_create_fake_l_geotiff_colormap, True, True, ""),
        (_create_fake_rgb_geotiff, False, True, ""),
        (_create_fake_rgb_geotiff, True, True, ""),
    ],
)
@pytest.mark.parametrize("colormap", [REDS_SPREAD_CMAP, REDS_MIN_CMAP])
@mock.patch("polar2grid.add_coastlines.ContourWriterAGG.add_overlay_from_dict")
def test_add_coastlines_basic(
    add_overlay_mock, tmp_path, gen_func, include_scale_offset, include_cmap_tag, output_ext, colormap
):
    from polar2grid.add_coastlines import main

    is_rgb = "rgb" in gen_func.__name__
    has_colormap = "colormap" in gen_func.__name__
    has_colors = colormap is not None and (has_colormap or is_rgb)

    fp = str(tmp_path / "test.tif")
    output_fp = fp.replace(".tif", ".png")
    gen_func(fp, colormap, include_scale_offset=include_scale_offset, include_colormap_tag=include_cmap_tag)
    extra_args = []
    if output_ext:
        output_fp = fp.replace(".tif", f"_new.{output_ext}")
        extra_args.extend(["-o", output_fp])

    with mocked_pydecorate_add_scale() as add_scale_mock:
        ret = main(["--add-coastlines", "--add-colorbar", fp] + extra_args)

    assert ret in [None, 0]
    assert os.path.isfile(output_fp)
    add_overlay_mock.assert_called_once()
    assert "coasts" in add_overlay_mock.call_args.args[0]
    add_scale_mock.assert_called_once()
    passed_cmap = add_scale_mock.call_args.kwargs["colormap"]
    _check_used_colormap(passed_cmap, has_colors, include_cmap_tag, include_scale_offset)

    with Image.open(output_fp) as img:
        img.load()
        arr = np.asarray(img)
        out_tags = dict(img.tag_v2) if output_fp.endswith(".tif") else {}
    # bottom of the image is a colorbar
    image_arr = arr[:940]
    _check_exp_image_colors(image_arr, colormap, 0, has_colors)
    _check_exp_image_colors(image_arr, colormap, 1, has_colors)
    _check_exp_image_colors(image_arr, colormap, 2, has_colors)
    assert (arr[940:] != 0).any()

    if output_fp.endswith(".tif"):
        with Image.open(fp) as in_img:
            in_img.load()
            in_tags = dict(in_img.tag_v2)
        assert len(out_tags) >= 14
        for key, val in out_tags.items():
            if key < 30000:
                continue
            assert in_tags[key] == val


@mock.patch("polar2grid.add_coastlines.ContourWriterAGG.add_overlay_from_dict")
def test_add_coastlines_multiple_inputs(add_overlay_mock, tmp_path):
    from polar2grid.add_coastlines import main

    colormap = REDS_MIN_CMAP
    fp1 = str(tmp_path / "test1.tif")
    _create_fake_l_geotiff_colormap(fp1, colormap, include_scale_offset=True, include_colormap_tag=True)
    fp2 = str(tmp_path / "test2.tif")
    _create_fake_l_geotiff_colormap(fp2, colormap, include_scale_offset=True, include_colormap_tag=True)
    extra_args = ["--colorbar-minor-tick-marks", "5.0", "--colorbar-tick-marks", "15.0"]

    with mocked_pydecorate_add_scale() as add_scale_mock:
        ret = main(["--add-coastlines", "--add-colorbar", fp1, fp2] + extra_args)

    assert ret in [None, 0]
    assert os.path.isfile(tmp_path / "test1.png")
    assert os.path.isfile(tmp_path / "test2.png")
    assert add_overlay_mock.call_count == 2
    assert "coasts" in add_overlay_mock.call_args.args[0]
    assert add_scale_mock.call_count == 2
    passed_cmap = add_scale_mock.call_args.kwargs["colormap"]
    _check_used_colormap(passed_cmap, True, True, True)

    for out_fn in ("test1.png", "test2.png"):
        img = Image.open(tmp_path / out_fn)
        arr = np.asarray(img)
        # bottom of the image is a colorbar
        image_arr = arr[:940]
        _check_exp_image_colors(image_arr, colormap, 0, True)
        _check_exp_image_colors(image_arr, colormap, 1, True)
        _check_exp_image_colors(image_arr, colormap, 2, True)
        assert (arr[940:] != 0).any()


def _check_used_colormap(passed_cmap, has_colors, include_cmap_tag, include_scale_offset):
    cmin = passed_cmap.values[0]
    cmax = passed_cmap.values[-1]
    cmap_size = passed_cmap.values.size
    exp_cmap_size = 255 if not has_colors or not include_cmap_tag else 3
    exp_cmin = 0.0
    exp_cmax = 255.0 if not has_colors or not include_scale_offset else 127.5
    assert cmap_size == exp_cmap_size
    assert cmin == exp_cmin
    assert cmax == exp_cmax

    if not has_colors:
        # no colormap, all black default colormap
        np.testing.assert_allclose(passed_cmap.colors[:, :3], 0)


def _check_exp_image_colors(image_arr, colormap, color_idx, has_colors):
    exp_raw_values = list(colormap.keys())
    cmap_colors = list(set(color[color_idx] for color in colormap.values()))
    exp_colors = cmap_colors if has_colors else exp_raw_values
    r_uniques = np.unique(image_arr[:, :, color_idx])
    np.testing.assert_allclose(r_uniques, exp_colors)


@contextlib.contextmanager
def mocked_pydecorate_add_scale():
    from polar2grid.add_coastlines import DecoratorAGG

    add_scale_mock = mock.Mock()
    decorator_mock = mock.Mock()

    def _create_decorator(img):
        dec = DecoratorAGG(img)
        add_scale_mock.side_effect = dec.add_scale
        dec.add_scale = add_scale_mock
        return dec

    decorator_mock.side_effect = _create_decorator
    with mock.patch("polar2grid.add_coastlines.DecoratorAGG", decorator_mock):
        yield add_scale_mock


@mock.patch("polar2grid.add_coastlines.ContourWriterAGG.add_overlay_from_dict")
def test_add_coastlines_bad_output_filenames(add_overlay_mock, tmp_path):
    from polar2grid.add_coastlines import main

    fp = str(tmp_path / "test.tif")
    _create_fake_l_geotiff(fp)
    extra_args = ["-o", "test1.png", "test2.png"]
    ret = main(["--add-coastlines", "--add-colorbar", fp] + extra_args)
    assert ret == -1
    assert not os.path.isfile(tmp_path / "test.png")
    add_overlay_mock.assert_not_called()



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Tests for polar2grid configuration files."""

import os
from glob import glob

import yaml


def pytest_generate_tests(metafunc):
    """Generate parametrized tests to run on all YAML files.

    Automatically used by pytest as a hook..

    """
    if "yaml_config_file" in metafunc.fixturenames:
        root_dir = os.path.join(os.path.dirname(__file__), "..")
        glob_pat = os.path.join(root_dir, "etc", "**", "*.yaml")
        p2g_yaml_files = sorted(glob(glob_pat, recursive=True))
        assert len(p2g_yaml_files) != 0
        metafunc.parametrize("yaml_config_file", p2g_yaml_files)


def test_valid_yaml_files(yaml_config_file):
    """Test basic YAML syntax for all config files."""
    with open(yaml_config_file, "r") as yaml_file:
        yaml.load(yaml_file, Loader=yaml.UnsafeLoader)



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2019-2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Produce various text and image outputs for debugging input data."""

import logging
import sys

import dask
import dask.array as da
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from satpy import Scene

matplotlib.use("agg")

LOG = logging.getLogger(__name__)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Produce various debug images and text about input data.")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG (default INFO)",
    )
    parser.add_argument("reader", help="Satpy reader to use")
    parser.add_argument("product", help="Product to load")
    parser.add_argument("files", nargs="*", help="Files to load data from")
    args = parser.parse_args()

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)])

    dask.config.set(num_workers=4)
    scn = Scene(reader=args.reader, filenames=args.files)
    scn.load([args.product])
    data = scn[args.product].compute()
    start_time = data.attrs["start_time"]
    lons, lats = da.compute(*data.attrs["area"].get_lonlats())

    print("## Longitude:")
    print("   Minimum: ", np.nanmin(lons))
    print("   Maximum: ", np.nanmax(lons))
    print("   Has NaNs: ", np.isnan(lons).any())
    print("   Shape: ", lons.shape)
    print("   Data Type: ", lons.dtype)
    fig, ax = plt.subplots()
    img = ax.imshow(lons, vmin=-180, vmax=180)
    cbar = fig.colorbar(img)
    ax.set_title("Longitude")
    cbar.set_label("Degrees")
    fig.savefig("longitude_{:%Y%m%d_%H%M%S}.png".format(start_time))

    print("## Latitude:")
    print("   Minimum: ", np.nanmin(lats))
    print("   Maximum: ", np.nanmax(lats))
    print("   Has NaNs: ", np.isnan(lats).any())
    print("   Shape: ", lats.shape)
    print("   Data Type: ", lats.dtype)
    fig, ax = plt.subplots()
    img = ax.imshow(lats, vmin=-90, vmax=90)
    cbar = fig.colorbar(img)
    ax.set_title("Latitude")
    cbar.set_label("Degrees")
    fig.savefig("latitude_{:%Y%m%d_%H%M%S}.png".format(start_time))

    print("## {}:".format(args.product))
    print("   Minimum: ", np.nanmin(data))
    print("   Maximum: ", np.nanmax(data))
    print("   Has NaNs: ", np.isnan(data).values.any().item())
    print("   Shape: ", data.shape)
    print("   Data Type: ", data.dtype)
    print("   Attributes: ", "\n      ".join("{}: {}".format(k, v) for k, v in data.attrs.items()))
    fig, ax = plt.subplots()
    img = ax.imshow(data)
    cbar = fig.colorbar(img)
    ax.set_title(args.product)
    cbar.set_label(data.attrs.get("units", "<unknown>"))
    fig.savefig("{}_{:%Y%m%d_%H%M%S}.png".format(args.product, start_time))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Helper functions for filtering certain products."""

import logging
from typing import Dict, List, Optional, Union

from satpy import Scene

from ..utils.dynamic_imports import get_reader_attr
from .day_night import DayCoverageFilter, NightCoverageFilter

logger = logging.getLogger(__name__)


def _merge_filter_critera(*readers_criteria: Dict[str, Union[List[str], None]]):
    result = {}
    for reader_criteria in readers_criteria:
        if reader_criteria is None:
            continue
        for criteria_key, criteria_value in reader_criteria.items():
            if criteria_value is not None:
                result.setdefault(criteria_key, []).extend(criteria_value)
    if not result:
        # no criteria were provided/found
        return None
    return result


def _get_single_reader_filter_criteria(reader: str, filter_name: str):
    filter_criteria = get_reader_attr(reader, "FILTERS", {})
    return filter_criteria.get(filter_name)


def get_reader_filter_criteria(reader_names: List[str], filter_name: str):
    """Get reader configured filter information."""
    readers_criteria = [_get_single_reader_filter_criteria(reader, filter_name) for reader in reader_names]
    criteria = _merge_filter_critera(*readers_criteria)
    return criteria


def _filter_scene_day_only_products(
    input_scene: Scene, filter_criteria: dict, sza_threshold: float = 100.0, day_fraction: Optional[float] = None
):
    """Run filtering for products that need a certain amount of day data."""
    if day_fraction is None:
        day_fraction = 0.1
    logger.info("Running day coverage filtering...")
    day_filter = DayCoverageFilter(filter_criteria, sza_threshold=sza_threshold, day_fraction=day_fraction)
    return day_filter.filter_scene(input_scene)


def _filter_scene_night_only_products(
    input_scene: Scene, filter_criteria: dict, sza_threshold: float = 100.0, night_fraction: Optional[float] = None
):
    """Run filtering for products that need a certain amount of day data."""
    if night_fraction is None:
        night_fraction = 0.1
    logger.info("Running night coverage filtering...")
    night_filter = NightCoverageFilter(filter_criteria, sza_threshold=sza_threshold, night_fraction=night_fraction)
    return night_filter.filter_scene(input_scene)


def filter_scene(
    input_scene: Scene,
    reader_names: List[str],
    sza_threshold: float = 100.0,
    day_fraction: Optional[float] = None,
    night_fraction: Optional[float] = None,
):
    if input_scene is not None and day_fraction is not False:
        criteria = get_reader_filter_criteria(reader_names, "day_only")
        input_scene = _filter_scene_day_only_products(
            input_scene,
            criteria,
            sza_threshold=sza_threshold,
            day_fraction=day_fraction,
        )
    if input_scene is not None and night_fraction is not False:
        criteria = get_reader_filter_criteria(reader_names, "night_only")
        input_scene = _filter_scene_night_only_products(
            input_scene,
            criteria,
            sza_threshold=sza_threshold,
            night_fraction=night_fraction,
        )
    return input_scene



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The Binary writer writes band data to a flat binary file.

By default it enhances the data based on the enhancement configuration file
and then saves the data to a flat binary file. The output data type will match
the input data for integer types (ex. uint8 -> uint8), but floating point types
will always be forced to 32-bit floats for consistency between readers and
changes in the low-level Satpy library. A different output type can be
specified using the ``--dtype`` flag. To turn off scaling of the data (a.k.a.
enhancements) the ``--no-enhance`` command line flag can be specified to write
the "raw" band data.

.. warning::

   Starting in Polar2Grid 3.2.0, output filename "data_type" fields now follow numpy-style naming.
   For example, 32-bit floats are now "float32" instead of "real4".
   Unsigned 8-bit integers are now "uint8" instead of "uint1".
   The dtype passed to the ``--dtype`` flag is unchanged (numpy style).

"""

from __future__ import annotations

import logging

import dask.array as da
import numpy as np
import xarray as xr
from satpy.writers.core.image import ImageWriter
from satpy.enhancements.enhancer import get_enhanced_image

from polar2grid.core.dtype import NUMPY_DTYPE_STRS, clip_to_data_type, dtype_to_str, int_or_float, str_to_dtype
from polar2grid.core.script_utils import NumpyDtypeList
from polar2grid.utils.legacy_compat import convert_p2g_pattern_to_satpy

logger = logging.getLogger(__name__)

DEFAULT_OUTPUT_FILENAMES = {
    "polar2grid": {
        None: "{platform_name!l}_{sensor!l}_{p2g_name}_{start_time:%Y%m%d_%H%M%S}_{area.area_id}.dat",
    },
    "geo2grid": {
        None: "{platform_name!u}_{sensor!u}_{p2g_name}_{start_time:%Y%m%d_%H%M%S}_{area.area_id}.dat",
        "abi_l1b": "{platform_name!u}_{sensor!u}_{observation_type}{scene_abbr}_"
        "{p2g_name}_{start_time:%Y%m%d_%H%M%S}_{area.area_id}.dat",
        "clavrx": "{platform_name!l}_{sensor!l}_{p2g_name}_{start_time:%Y%m%d_%H%M%S}_{area.area_id}.dat",
    },
}


class FlatBinaryWriter(ImageWriter):
    """Write data to disk as flat binary files."""

    def save_dataset(
        self, dataset, filename=None, fill_value=None, overlay=None, decorate=None, compute=True, **kwargs
    ):
        """Save the ``dataset`` to a given ``filename``.

        This method creates an enhanced image using :func:`get_enhanced_image`.
        The image is then passed to :meth:`save_image`. See both of these
        functions for more details on the arguments passed to this method.

        """
        img = get_enhanced_image(
            dataset.squeeze(),
            enhance=self.enhancer,
            overlay=overlay,
            decorate=decorate,
        )
        kwargs["dtype"] = kwargs.get("dtype") or self._get_default_dtype(dataset)
        return self.save_image(img, filename=filename, compute=compute, fill_value=fill_value, **kwargs)

    @staticmethod
    def _get_default_dtype(data_arr: xr.DataArray) -> np.dtype:
        if data_arr.dtype == np.float64:
            return np.float32
        return data_arr.dtype

    def save_image(self, img, filename=None, compute=True, dtype=None, fill_value=None, **kwargs):
        filename = filename or self.get_filename(
            data_type=dtype_to_str(dtype), rows=img.data.shape[0], columns=img.data.shape[1], **img.data.attrs
        )

        data = self._prep_data(img.data, dtype, fill_value)

        logger.info("Saving product %s to binary file %s", img.data.attrs["p2g_name"], filename)
        dst = np.memmap(filename, shape=img.data.shape, dtype=dtype, mode="w+")
        if compute:
            da.store(data, dst)
            return filename
        return [data], [dst]

    def _prep_data(self, data: xr.DataArray, dtype: np.dtype, fill_value) -> da.Array:
        fill = data.attrs.get("_FillValue", np.nan)
        if fill_value is None:
            fill_value = fill
        final_data = data.data
        if self.enhancer and np.issubdtype(data.dtype, np.floating) and not np.issubdtype(dtype, np.floating):
            # going from float -> int and the data was enhanced
            # scale the data to fit the integer dtype
            rmin, rmax = np.iinfo(dtype).min, np.iinfo(dtype).max
            final_data = final_data * (rmax - rmin) + rmin
        final_data = clip_to_data_type(final_data, dtype)

        same_fill = np.isnan(fill) and np.isnan(fill_value) or fill == fill_value
        if data.dtype == dtype and same_fill:
            return final_data

        final_data = final_data.astype(dtype)
        if same_fill:
            return final_data

        fill_mask = np.isnan(final_data) if np.isnan(fill) else final_data == fill
        final_data = da.where(fill_mask, fill_value, final_data)
        return final_data


def add_writer_argument_groups(parser, group=None):
    if group is None:
        group = parser.add_argument_group(title="Binary Writer")

    group.add_argument(
        "--output-filename",
        dest="filename",
        type=convert_p2g_pattern_to_satpy,
        help="Custom file pattern to save dataset to",
    )
    group.add_argument(
        "--dtype",
        choices=NumpyDtypeList(NUMPY_DTYPE_STRS),
        type=str_to_dtype,
        help="Data type of the output file (8-bit unsigned integer by default - uint8)",
    )
    group.add_argument(
        "--no-enhance",
        dest="enhance",
        action="store_false",
        help="Don't enhance the data before saving it",
    )
    group.add_argument(
        "--fill-value",
        dest="fill_value",
        type=int_or_float,
        help="Replace invalid values with the specified value. Floating-point "
        "products typically use NaN while integer fields will use 0 or "
        "the max value for that data type.",
    )
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2018 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The AHI High Rate Information Transmission (HRIT) Reader operates on
standard Japan Meteorological Agency (JMA) Himawari-8 and Himawari-9
Advanced Himawari Imager (AHI) HRIT Digital Video Broadcast (DVB)
HimawariCast files.  This broadcast consists of a subset of 14 bands
at reduced spatial resolution.  The AHI HRIT reader works off of
the input filenames to determine if a file is supported by Geo2Grid.
Files usually have the following naming scheme:

    IMG_DK01B04_201809100300

These are the mission compliant radiance file naming conventions
used by JMA. The AHI HRIT reader supports a subset of instrument spectral bands,
identified in Geo2Grid as the products shown in the table below. The
AHI HRIT reader can be provided to the main geo2grid.sh script
using the ``-r`` option and the reader name ``ahi_hrit``.

The list of supported products includes natural color imagery.
This is created by means of a python based atmospheric Rayleigh
scattering correction algorithm that is executed as part of the |project| AHI
HRIT reader, along with sharpening to the highest spatial resolution.

Please note that true color image RGB creation is not supported for
HimawariCast because AHI Band 1 (Blue) and Band 2 (Green) are not
included.

For more information on the creation of RGBs, please see the
:ref:`RGB section <getting_started_rgb>`.

+---------------------------+-----------------------------------------------------+
| **Product Name**          | **Description**                                     |
+===========================+=====================================================+
| B03                       | Channel 3 (0.64um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B04                       | Channel 4 (0.86um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B05                       | Channel 5 (1.6um) Reflectance Band                  |
+---------------------------+-----------------------------------------------------+
| B06                       | Channel 6 (2.3um) Reflectance Band                  |
+---------------------------+-----------------------------------------------------+
| B07                       | Channel 7 (3.9um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B08                       | Channel 8 (6.2um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B09                       | Channel 9 (6.9um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B10                       | Channel 10 (7.3um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B11                       | Channel 11 (8.6um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B12                       | Channel 12 (9.6um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B13                       | Channel 13 (10.4um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B14                       | Channel 14 (11.2um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B15                       | Channel 15 (12.4um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B16                       | Channel 16 (13.3um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| natural_color             | Ratio sharpened rayleigh corrected natural color    |
+---------------------------+-----------------------------------------------------+
| airmass                   | Air mass RGB                                        |
+---------------------------+-----------------------------------------------------+
| ash                       | Ash RGB                                             |
+---------------------------+-----------------------------------------------------+
| dust                      | Dust RGB                                            |
+---------------------------+-----------------------------------------------------+
| fog                       | Fog RGB                                             |
+---------------------------+-----------------------------------------------------+
| night_microphysics        | Night Microphysics RGB                              |
+---------------------------+-----------------------------------------------------+

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from ._base import ReaderProxyBase

READER_PRODUCTS = ["B{:02d}".format(x) for x in range(3, 17)]
COMPOSITE_PRODUCTS = [
    "natural_color",
    "airmass",
    "ash",
    "dust",
    "fog",
    "night_microphysics",
]


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_geo2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS[:1]

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="AHI HRIT Reader")
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2020 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The VIIRS SDR Reader operates on Sensor Data Record (SDR) HDF5 files from
the Suomi National Polar-orbiting Partnership (SNPP), NOAA21 and  NOAA20
Visible/Infrared Imager Radiometer Suite (VIIRS) instruments. The VIIRS
SDR reader requires filenames to match one of a couple different standard
filename schemes used for official products. SDR files are typically named
as below and have corresponding geolocation files::

    SVI01_j02_d20260205_t1911178_e1912424_b16791_c20260205194106784585_cspp_dev.h5

The VIIRS SDR reader supports all instrument spectral bands, identified as
the products shown below. It supports terrain corrected or non-terrain corrected
navigation files. Geolocation files must be included when specifying filepaths to
readers and ``polar2grid.sh``. The VIIRS reader can be specified to the ``polar2grid.sh`` script
with the reader name ``viirs_sdr``.

This reader's default remapping algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--weight-delta-max`` parameter set to 40 and the
``--weight-distance-max`` parameter set to 2.

+---------------------------+-----------------------------------------------------+
| Product Name              | Description                                         |
+===========================+=====================================================+
| i01                       | I01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i02                       | I02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i03                       | I03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i04                       | I04 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| i05                       | I05 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| i01_rad                   | I01 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i02_rad                   | I02 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i03_rad                   | I03 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i04_rad                   | I04 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i05_rad                   | I05 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m01                       | M01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m02                       | M02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m03                       | M03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m04                       | M04 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m05                       | M05 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m06                       | M06 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m07                       | M07 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m08                       | M08 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m09                       | M09 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m10                       | M10 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m11                       | M11 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m12                       | M12 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m13                       | M13 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m14                       | M14 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m15                       | M15 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m16                       | M16 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m01_rad                   | M01 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m02_rad                   | M02 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m03_rad                   | M03 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m04_rad                   | M04 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m05_rad                   | M05 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m06_rad                   | M06 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m07_rad                   | M07 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m08_rad                   | M08 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m09_rad                   | M09 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m10_rad                   | M10 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m11_rad                   | M11 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m12_rad                   | M12 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m13_rad                   | M13 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m14_rad                   | M14 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m15_rad                   | M15 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m16_rad                   | M16 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| dnb                       | Raw DNB Band (not useful for images)                |
+---------------------------+-----------------------------------------------------+
| histogram_dnb             | Histogram Equalized DNB Band                        |
+---------------------------+-----------------------------------------------------+
| adaptive_dnb              | Adaptive Histogram Equalized DNB Band               |
+---------------------------+-----------------------------------------------------+
| dynamic_dnb               | Dynamic DNB Band from Steve Miller and              |
|                           | Curtis Seaman. Uses erf to scale the data.          |
+---------------------------+-----------------------------------------------------+
| hncc_dnb                  | Simplified High and Near-Constant Contrast          |
|                           | Approach from Stephan Zinke                         |
+---------------------------+-----------------------------------------------------+
| ifog                      | Temperature difference between I05 and I04          |
+---------------------------+-----------------------------------------------------+
| i_solar_zenith_angle      | I Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| i_solar_azimuth_angle     | I Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| i_sat_zenith_angle        | I Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| i_sat_azimuth_angle       | I Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| m_solar_zenith_angle      | M Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| m_solar_azimuth_angle     | M Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| m_sat_zenith_angle        | M Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| m_sat_azimuth_angle       | M Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| dnb_solar_zenith_angle    | DNB Band Solar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_solar_azimuth_angle   | DNB Band Solar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| dnb_sat_zenith_angle      | DNB Band Satellite Zenith Angle                     |
+---------------------------+-----------------------------------------------------+
| dnb_sat_azimuth_angle     | DNB Band Satellite Azimuth Angle                    |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_zenith_angle    | DNB Band Lunar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_azimuth_angle   | DNB Band Lunar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| true_color                | Ratio sharpened rayleigh corrected true color       |
+---------------------------+-----------------------------------------------------+
| false_color               | Ratio sharpened rayleigh corrected false color      |
+---------------------------+-----------------------------------------------------+

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery, Scene

from polar2grid.core.script_utils import ExtendConstAction

from ._base import ReaderProxyBase

PREFERRED_CHUNK_SIZE: int = 6400

I_VIS_PRODUCTS = [
    "I01",
    "I02",
    "I03",
]
I_IR_PRODUCTS = [
    "I04",
    "I05",
]
I_PRODUCTS = I_VIS_PRODUCTS + I_IR_PRODUCTS
I_ANGLE_PRODUCTS = [
    "i_solar_zenith_angle",
    "i_solar_azimuth_angle",
    "i_sat_zenith_angle",
    "i_sat_azimuth_angle",
]
M_VIS_PRODUCTS = [
    "M01",
    "M02",
    "M03",
    "M04",
    "M05",
    "M06",
    "M07",
    "M08",
    "M09",
    "M10",
    "M11",
]
M_IR_PRODUCTS = [
    "M12",
    "M13",
    "M14",
    "M15",
    "M16",
]
M_PRODUCTS = M_VIS_PRODUCTS + M_IR_PRODUCTS
M_ANGLE_PRODUCTS = [
    "m_solar_zenith_angle",
    "m_solar_azimuth_angle",
    "m_sat_zenith_angle",
    "m_sat_azimuth_angle",
]
DNB_PRODUCTS = [
    "histogram_dnb",
    "adaptive_dnb",
    "dynamic_dnb",
    "hncc_dnb",
]
DNB_ANGLE_PRODUCTS = [
    "dnb_solar_zenith_angle",
    "dnb_solar_azimuth_angle",
    "dnb_sat_zenith_angle",
    "dnb_sat_azimuth_angle",
    "dnb_lunar_zenith_angle",
    "dnb_lunar_azimuth_angle",
]
TRUE_COLOR_PRODUCTS = ["true_color"]
FALSE_COLOR_PRODUCTS = ["false_color"]
OTHER_COMPS = [
    "ifog",
]


PRODUCT_ALIASES = {}


def _process_legacy_and_rad_products(satpy_names, band_aliases, rad_aliases):
    """Map all lowercase band names to uppercase names and add radiance product."""
    for band in satpy_names:
        # P2G name is lowercase, Satpy is uppercase
        PRODUCT_ALIASES[band.lower()] = band
        band_aliases.append(band.lower())

        # radiance products for M and I bands
        rad_name = band.lower() + "_rad"
        dq = DataQuery(name=band, calibration="radiance")
        PRODUCT_ALIASES[rad_name] = dq
        rad_aliases.append(rad_name)


I_ALIASES = []
I_RAD_PRODUCTS = []
_process_legacy_and_rad_products(I_PRODUCTS, I_ALIASES, I_RAD_PRODUCTS)
M_ALIASES = []
M_RAD_PRODUCTS = []
_process_legacy_and_rad_products(M_PRODUCTS, M_ALIASES, M_RAD_PRODUCTS)

_AWIPS_TRUE_COLOR = ["viirs_crefl08", "viirs_crefl04", "viirs_crefl03"]
_AWIPS_FALSE_COLOR = ["viirs_crefl07", "viirs_crefl09", "viirs_crefl08"]

PRODUCT_ALIASES["dnb_solar_zenith_angle"] = DataQuery(name="dnb_solar_zenith_angle")
PRODUCT_ALIASES["dnb_solar_azimuth_angle"] = DataQuery(name="dnb_solar_azimuth_angle")
PRODUCT_ALIASES["dnb_sat_zenith_angle"] = DataQuery(name="dnb_satellite_zenith_angle")
PRODUCT_ALIASES["dnb_sat_azimuth_angle"] = DataQuery(name="dnb_satellite_azimuth_angle")
PRODUCT_ALIASES["dnb_lunar_zenith_angle"] = DataQuery(name="dnb_lunar_zenith_angle")
PRODUCT_ALIASES["dnb_lunar_azimuth_angle"] = DataQuery(name="dnb_lunar_azimuth_angle")
PRODUCT_ALIASES["m_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=742)
PRODUCT_ALIASES["m_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=742)
PRODUCT_ALIASES["i_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=371)
PRODUCT_ALIASES["i_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=371)

DEFAULT_PRODUCTS = I_ALIASES + M_ALIASES + DNB_PRODUCTS[1:] + TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS + OTHER_COMPS
P2G_PRODUCTS = I_ALIASES + M_ALIASES + DNB_PRODUCTS + I_RAD_PRODUCTS + M_RAD_PRODUCTS
P2G_PRODUCTS += I_ANGLE_PRODUCTS + M_ANGLE_PRODUCTS + DNB_ANGLE_PRODUCTS + OTHER_COMPS
P2G_PRODUCTS += TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS
P2G_PRODUCTS += [
    "viirs_crefl01",
    "viirs_crefl02",
    "viirs_crefl03",
    "viirs_crefl04",
    "viirs_crefl07",
    "viirs_crefl08",
    "viirs_crefl09",
]
P2G_PRODUCTS += ["dynamic_dnb_saturation", "unsharp_true_color"]

FILTERS = {
    "day_only": {
        "standard_name": [
            "toa_bidirectional_reflectance",
            "true_color",
            "false_color",
            "natural_color",
            "corrected_reflectance",
        ],
    },
    "night_only": {
        "standard_name": ["temperature_difference"],
    },
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def __init__(self, scn: Scene, user_products: list[str]):
        self._modified_aliases = PRODUCT_ALIASES.copy()
        if "dynamic_dnb_saturation" in user_products:
            # they specified --dnb-saturation-correction
            # let's modify the aliases so dynamic_dnb points to this product
            user_products.remove("dynamic_dnb_saturation")
            self._modified_aliases["dynamic_dnb"] = DataQuery(name="dynamic_dnb_saturation")
        super().__init__(scn, user_products)

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_PRODUCTS

    def get_all_products(self):
        """Get all polar2grid products that could be loaded."""
        return P2G_PRODUCTS

    @property
    def _aliases(self):
        return self._modified_aliases


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="VIIRS SDR Reader")

    group.add_argument(
        "--i-bands",
        dest="products",
        action=ExtendConstAction,
        const=I_ALIASES,
        help="Add all I-band raw products to list of products",
    )
    group.add_argument(
        "--m-bands",
        dest="products",
        action=ExtendConstAction,
        const=M_ALIASES,
        help="Add all M-band raw products to list of products",
    )
    group.add_argument(
        "--dnb-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=DNB_ANGLE_PRODUCTS,
        help="Add DNB-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--dnb-saturation-correction",
        dest="products",
        action=ExtendConstAction,
        const=["dynamic_dnb_saturation"],
        help="Enable dynamic DNB saturation correction (normally used for aurora scenes)",
    )
    group.add_argument(
        "--i-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=I_ANGLE_PRODUCTS,
        help="Add I-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--m-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=M_ANGLE_PRODUCTS,
        help="Add M-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--m-rad-products",
        dest="products",
        action=ExtendConstAction,
        const=M_RAD_PRODUCTS,
        help="Add M-band geolocation radiance products to list of products",
    )
    group.add_argument(
        "--i-rad-products",
        dest="products",
        action=ExtendConstAction,
        const=I_RAD_PRODUCTS,
        help="Add I-band geolocation radiance products to list of products",
    )
    group.add_argument(
        "--awips-true-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_TRUE_COLOR,
        help="Add individual CREFL corrected products to create the 'true_color' composite in AWIPS.",
    )
    group.add_argument(
        "--awips-false-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_FALSE_COLOR,
        help="Add individual CREFL corrected products to create the 'false_color' composite in AWIPS.",
    )
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The CF writer puts gridded image data into a `CF-compliant` NetCDF file.

All datasets to be saved must have the same projection coordinates ``x`` and ``y``. If a scene holds datasets with
different grids, the CF compliant workaround is to save the datasets to separate files.

"""

import json
import logging

LOG = logging.getLogger(__name__)

# reader_name -> filename
DEFAULT_OUTPUT_FILENAMES = {
    "polar2grid": {
        None: "{platform_name}_{sensor}_{start_time:%Y%m%d_%H%M%S}.nc",
    },
    "geo2grid": {
        None: "{platform_name}_{sensor}_{start_time:%Y%m%d_%H%M%S}.nc",
    },
}


def add_writer_argument_groups(parser, group=None):
    if group is None:
        group = parser.add_argument_group(title="cf Writer")
    group.add_argument(
        "--output-filename",
        dest="filename",
        help="Custom file pattern to save dataset to",
    )
    group.add_argument(
        "--groups",
        type=json.loads,
        help="Group datasets according to the given assignment: `{'group_name': ['dataset1', 'dataset2', ...]}`.",
    )

    group.add_argument(
        "--header-attrs",
        type=json.loads,
        help="Global attributes to be included.",
    )

    group.add_argument(
        "--engine",
        default="netcdf4",
        help="Engine for writing NetCDF file \
             (options: '"
        "netcdf4"
        "', '"
        "scipy"
        "', '"
        "h5netcdf"
        "').",
    )
    group.add_argument("--epoch-units", dest="epoch", help="Reference unit for the netCDF time coordinates.")

    group.add_argument(
        "--exclude-attrs",
        nargs="+",
        help="List of dataset attributes to be excluded",
    )
    group.add_argument(
        "--include-lonlats",
        action="store_true",
        help="Include latitude and longitude coordinates.",
    )

    group.add_argument("--pretty", action="store_true", help="Do not prefix coordinate with corresponding dataset name")

    group.add_argument(
        "--include-orig-name",
        action="store_true",
        help="Include the original dataset name as a variable attribute in the final netcdf.",
    )

    group.add_argument(
        "--flatten-attrs",
        action="store_true",
        help="If invoked, flatten dict-type attributes",
    )

    group.add_argument(
        "--numeric-name-prefix",
        default="CHANNEL_",
        help="Prefix added to each variable. For name starting with a digit.Use '' or None to leave this out..",
    )

    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2014-2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Various utility functions and scripts."""



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2016 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
#     Written by David Hoese    March 2016
#     University of Wisconsin-Madison
#     Space Science and Engineering Center
#     1225 West Dayton Street
#     Madison, WI  53706
#     david.hoese@ssec.wisc.edu
"""The VIIRS Level 1B Reader operates on NASA Level 1B (L1B) NetCDF files.

The files are from the Visible/Infrared Imager Radiometer Suite (VIIRS)
instrument. The VIIRS L1B reader analyzes the user provided filenames
to determine if a file can be used. Files usually have the following
naming scheme (example from the SNPP VIIRS instrument)::

    VL1BI_snpp_d20160101_t185400_c20160301041812.nc

The VIIRS L1B reader supports all instrument spectral bands, identified as
the products shown below.  Geolocation files must be included when
specifying filepaths to readers and ``polar2grid.sh``.  Therefore, the
creation of The VIIRS L1B frontend can be specified to the Polar2Grid
glue script with the frontend name ``viirs_l1b``.

The list of supported products includes true and false color imagery.
These are created  by means of a python based atmospheric Rayleigh
Scattering algorithm that is executed as part of the Polar2Grid VIIRS L1B
reader.

.. note::

    The VIIRS L1B reader only supports the NASA L1B version 2.0 file format.
    Previous and future versions may work for some products, but are not
    guaranteed.

This reader's default resampling algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--weight-delta-max`` parameter is set to 40 and the
``--weight-distance-max`` parameter is set to 2.

+---------------------------+-----------------------------------------------------+
| Product Name              | Description                                         |
+===========================+=====================================================+
| i01                       | I01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i02                       | I02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i03                       | I03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i04                       | I04 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| i05                       | I05 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m01                       | M01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m02                       | M02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m03                       | M03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m04                       | M04 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m05                       | M05 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m06                       | M06 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m07                       | M07 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m08                       | M08 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m09                       | M09 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m10                       | M10 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m11                       | M11 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m12                       | M12 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m13                       | M13 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m14                       | M14 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m15                       | M15 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m16                       | M16 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| DNB                       | Raw DNB Band (not useful for images)                |
+---------------------------+-----------------------------------------------------+
| histogram_dnb             | Histogram Equalized DNB Band                        |
+---------------------------+-----------------------------------------------------+
| adaptive_dnb              | Adaptive Histogram Equalized DNB Band               |
+---------------------------+-----------------------------------------------------+
| dynamic_dnb               | Dynamic DNB Band from Steve Miller and              |
|                           | Curtis Seaman. Uses erf to scale the data           |
+---------------------------+-----------------------------------------------------+
| hncc_dnb                  | Simplified High and Near-Constant Contrast          |
|                           | Approach from Stephan Zinke                         |
+---------------------------+-----------------------------------------------------+
| ifog                      | Temperature difference between I05 and I04          |
+---------------------------+-----------------------------------------------------+
| i_solar_zenith_angle      | I Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| i_solar_azimuth_angle     | I Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| i_sat_zenith_angle        | I Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| i_sat_azimuth_angle       | I Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| m_solar_zenith_angle      | M Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| m_solar_azimuth_angle     | M Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| m_sat_zenith_angle        | M Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| m_sat_azimuth_angle       | M Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| dnb_solar_zenith_angle    | DNB Band Solar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_solar_azimuth_angle   | DNB Band Solar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| dnb_sat_zenith_angle      | DNB Band Satellite Zenith Angle                     |
+---------------------------+-----------------------------------------------------+
| dnb_sat_azimuth_angle     | DNB Band Satellite Azimuth Angle                    |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_zenith_angle    | DNB Band Lunar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_azimuth_angle   | DNB Band Lunar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| true_color                | Ratio sharpened rayleigh corrected true color       |
+---------------------------+-----------------------------------------------------+
| false_color               | Ratio sharpened rayleigh corrected false color      |
+---------------------------+-----------------------------------------------------+


For reflectance/visible products a check is done to make sure that at least
10% of the swath is day time. Data is considered day time where solar zenith
angle is less than 100 degrees.

"""

from __future__ import annotations

__docformat__ = "restructuredtext en"

import logging
from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery, Scene

from polar2grid.core.script_utils import ExtendConstAction

from ._base import ReaderProxyBase

logger = logging.getLogger(__name__)

I_VIS_PRODUCTS = [
    "I01",
    "I02",
    "I03",
]
I_IR_PRODUCTS = [
    "I04",
    "I05",
]
I_PRODUCTS = I_VIS_PRODUCTS + I_IR_PRODUCTS
I_ANGLE_PRODUCTS = [
    "i_solar_zenith_angle",
    "i_solar_azimuth_angle",
    "i_sat_zenith_angle",
    "i_sat_azimuth_angle",
]
M_VIS_PRODUCTS = [
    "M01",
    "M02",
    "M03",
    "M04",
    "M05",
    "M06",
    "M07",
    "M08",
    "M09",
    "M10",
    "M11",
]
M_IR_PRODUCTS = [
    "M12",
    "M13",
    "M14",
    "M15",
    "M16",
]
M_PRODUCTS = M_VIS_PRODUCTS + M_IR_PRODUCTS

M_ANGLE_PRODUCTS = [
    "m_solar_zenith_angle",
    "m_solar_azimuth_angle",
    "m_sat_zenith_angle",
    "m_sat_azimuth_angle",
]
DNB_PRODUCTS = [
    "histogram_dnb",
    "adaptive_dnb",
    "dynamic_dnb",
    "hncc_dnb",
]

DNB_ANGLE_PRODUCTS = [
    "dnb_solar_zenith_angle",
    "dnb_solar_azimuth_angle",
    "dnb_sat_zenith_angle",
    "dnb_sat_azimuth_angle",
    "dnb_lunar_zenith_angle",
    "dnb_lunar_azimuth_angle",
]

TRUE_COLOR_PRODUCTS = ["true_color"]
FALSE_COLOR_PRODUCTS = ["false_color"]
OTHER_COMPS = [
    "ifog",
]

PRODUCT_ALIASES = {}


def _process_legacy_products(satpy_names, band_aliases):
    """Map all lowercase band names to uppercase names and add radiance product."""
    for band in satpy_names:
        query = band
        # P2G name is lowercase, Satpy is uppercase
        if band in I_VIS_PRODUCTS:
            query = DataQuery(name=band, modifiers=("sunz_corrected_iband",))
        if band in M_VIS_PRODUCTS:
            query = DataQuery(name=band, modifiers=("sunz_corrected",))
        PRODUCT_ALIASES[band.lower()] = query
        band_aliases.append(band.lower())


I_ALIASES = []
_process_legacy_products(I_PRODUCTS, I_ALIASES)
M_ALIASES = []
_process_legacy_products(M_PRODUCTS, M_ALIASES)

_AWIPS_TRUE_COLOR = ["viirs_crefl08", "viirs_crefl04", "viirs_crefl03"]
_AWIPS_FALSE_COLOR = ["viirs_crefl07", "viirs_crefl09", "viirs_crefl08"]

PRODUCT_ALIASES["dnb_solar_zenith_angle"] = DataQuery(name="dnb_solar_zenith_angle")
PRODUCT_ALIASES["dnb_solar_azimuth_angle"] = DataQuery(name="dnb_solar_azimuth_angle")
PRODUCT_ALIASES["dnb_lunar_zenith_angle"] = DataQuery(name="dnb_lunar_zenith_angle")
PRODUCT_ALIASES["dnb_lunar_azimuth_angle"] = DataQuery(name="dnb_lunar_azimuth_angle")
PRODUCT_ALIASES["m_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=742)
PRODUCT_ALIASES["m_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=742)
PRODUCT_ALIASES["solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=742)
PRODUCT_ALIASES["solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=742)
PRODUCT_ALIASES["sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=742)
PRODUCT_ALIASES["sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=742)
PRODUCT_ALIASES["i_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=371)
PRODUCT_ALIASES["i_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=371)
# old "satellite" name
PRODUCT_ALIASES["satellite_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=742)
PRODUCT_ALIASES["satellite_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=742)
PRODUCT_ALIASES["i_satellite_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_satellite_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=371)

DEFAULT_PRODUCTS = I_ALIASES + M_ALIASES + TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS + DNB_PRODUCTS[1:] + OTHER_COMPS

P2G_PRODUCTS = I_ALIASES + M_ALIASES + TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS + DNB_PRODUCTS
P2G_PRODUCTS += DNB_ANGLE_PRODUCTS + M_ANGLE_PRODUCTS + I_ANGLE_PRODUCTS + OTHER_COMPS

FILTERS = {
    "day_only": {
        "standard_name": [
            "toa_bidirectional_reflectance",
            "true_color",
            "false_color",
            "natural_color",
            "corrected_reflectance",
        ],
    },
    "night_only": {
        "standard_name": ["temperature_difference"],
    },
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def __init__(self, scn: Scene, user_products: list[str]):
        self._modified_aliases = PRODUCT_ALIASES.copy()

        try:
            # they specified --normalized-radiances
            user_products.remove("normalized_radiances")
            apply_sunz = False
        except ValueError:
            apply_sunz = True

        i_modifiers = ("sunz_corrected_iband",) if apply_sunz else ()
        if i_modifiers:
            logger.debug(f"Using I-band visible channel modifiers: {i_modifiers}")
        for vis_product in I_VIS_PRODUCTS:
            self._modified_aliases[vis_product] = DataQuery(name=vis_product, modifiers=i_modifiers)
            self._modified_aliases[vis_product.lower()] = DataQuery(name=vis_product, modifiers=i_modifiers)

        m_modifiers = ("sunz_corrected",) if apply_sunz else ()
        if m_modifiers:
            logger.debug(f"Using M-band visible channel modifiers: {m_modifiers}")
        for vis_product in M_VIS_PRODUCTS:
            self._modified_aliases[vis_product] = DataQuery(name=vis_product, modifiers=m_modifiers)
            self._modified_aliases[vis_product.lower()] = DataQuery(name=vis_product, modifiers=m_modifiers)
        super().__init__(scn, user_products)

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_PRODUCTS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return P2G_PRODUCTS

    @property
    def _aliases(self) -> dict[str, DataQuery]:
        return self._modified_aliases


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="VIIRS l1b Reader")
    group.add_argument(
        "--i-bands",
        dest="products",
        action=ExtendConstAction,
        const=I_ALIASES,
        help="Add all I-band raw products to list of products",
    )
    group.add_argument(
        "--m-bands",
        dest="products",
        action=ExtendConstAction,
        const=M_ALIASES,
        help="Add all M-band raw products to list of products",
    )
    group.add_argument(
        "--normalized-radiances",
        dest="products",
        action=ExtendConstAction,
        const=["normalized_radiances"],
        help="Do not apply '/ cos(SZA)' when loading visible bands.",
    )
    group.add_argument(
        "--awips-true-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_TRUE_COLOR,
        help="Add the True Color product to the list of products",
    )
    group.add_argument(
        "--awips-false-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_FALSE_COLOR,
        help="Add the False Color product to the list of products",
    )
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2015-2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Test grid manager."""

from pyresample import AreaDefinition, DynamicAreaDefinition

from polar2grid.grids import GridManager


def test_grid_manager_basic(builtin_test_grids_conf, viirs_sdr_i_swath_def):
    """Test basic parsing of .conf files."""
    gm = GridManager(*builtin_test_grids_conf)
    for grid_name in gm.grid_information:
        grid_def = gm.get_grid_definition(grid_name)
        area_def = grid_def.to_satpy_area()
        assert isinstance(area_def, (AreaDefinition, DynamicAreaDefinition))
        if isinstance(area_def, DynamicAreaDefinition):
            new_area_def = area_def.freeze(viirs_sdr_i_swath_def)
            assert new_area_def.shape[0] > 0
            assert new_area_def.shape[1] > 0
            assert new_area_def.area_extent[0] < new_area_def.area_extent[2]
            assert new_area_def.area_extent[1] < new_area_def.area_extent[3]



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2017 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
#     Written by David Hoese    June 2017
#     University of Wisconsin-Madison
#     Space Science and Engineering Center
#     1225 West Dayton Street
#     Madison, WI  53706
#     david.hoese@ssec.wisc.edu
"""The CLAVR-x reader is for reading files created by the Community
Satellite Processing Package (CSPP) Clouds from AVHRR Extended
(CLAVR-x) processing system software. The CLAVR-x reader
supports CSPP CLAVR-x product files created from VIIRS, MODIS
and AVHRR imaging sensors in the native HDF4 format. For more
information on this product, please visit the CSPP LEO
website: `https://cimss.ssec.wisc.edu/cspp/`.

The reader can be specified with the ``polar2grid.sh``
command using the ``clavrx`` reader name.
The CLAVR-x reader provides the following products, which include
support for the VIIRS Day/Night Band Lunar Reflectance:

+------------------------+---------------------------------------------+
| Product Name           | Description                                 |
+========================+=============================================+
| cloud_type             | Cloud Type                                  |
+------------------------+---------------------------------------------+
| cld_height_acha        | Cloud Top Height (m)                        |
+------------------------+---------------------------------------------+
| cld_temp_acha          | Cloud Top Temperature (K)                   |
+------------------------+---------------------------------------------+
| cld_emiss_acha         | Cloud Emissivity                            |
+------------------------+---------------------------------------------+
| cld_opd_dcomp          | Cloud Optical Depth Daytime                 |
+------------------------+---------------------------------------------+
| cld_opd_nlcomp         | Cloud Optical Depth Nighttime               |
+------------------------+---------------------------------------------+
| cld_reff_dcomp         | Cloud Effective Radius Daytime (micron)     |
+------------------------+---------------------------------------------+
| cld_reff_nlomp         | Cloud Effective Radius Nighttime (micron)   |
+------------------------+---------------------------------------------+
| cloud_phase            | Cloud Phase  (5 categories)                 |
+------------------------+---------------------------------------------+
| rain_rate              | Rain Rate (mm/hr)                           |
+------------------------+---------------------------------------------+
| refl_lunar_dnb_nom     | Lunar Reflectance (VIIRS DNB nighttime only)|
+------------------------+---------------------------------------------+

"""

from __future__ import annotations

import os
from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from ._base import ReaderProxyBase

# Limit the number of products shown to Polar2Grid users
# if the user uses the environment variable they can display more
ADVERTISED_DATASETS = os.environ.get("CLAVRX_AD_DATASETS", None)
if ADVERTISED_DATASETS == "all":
    ADVERTISED_DATASETS = None
elif ADVERTISED_DATASETS:
    ADVERTISED_DATASETS = ADVERTISED_DATASETS.split(" ")
else:
    ADVERTISED_DATASETS = {
        "cloud_type",
        "cld_temp_acha",
        "cld_height_acha",
        "cloud_phase",
        "cld_opd_dcomp",
        "cld_opd_nlcomp",
        "cld_reff_dcomp",
        "cld_reff_nlcomp",
        "cld_emiss_acha",
        "refl_lunar_dnb_nom",
        "rain_rate",
    }

DEFAULT_DATASETS = [
    "cloud_type",
    "cld_temp_acha",
    "cld_height_acha",
    "cloud_phase",
    "cld_opd_dcomp",
    "cld_opd_nlcomp",
    "cld_reff_dcomp",
    "cld_reff_nlcomp",
    "cld_emiss_acha",
    "refl_lunar_dnb_nom",
    "rain_rate",
]

FILTERS = {
    "day_only": {
        "name": [
            "cld_opd_dcomp",
            "cld_reff_dcomp",
        ],
    },
    "night_only": {
        "name": [
            "cld_opd_nlcomp",
            "cld_reff_nlcomp",
            "refl_lunar_dnb_nom",
        ],
    },
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_geo2grid_reader = True
    is_polar2grid_reader = True

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return ADVERTISED_DATASETS

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_DATASETS


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser."""
    if group is None:
        group = parser.add_argument_group(title="CLAVR-x Reader")
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2014-2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Script for comparing writer output."""

from __future__ import annotations

import logging
import os
import sys
from dataclasses import dataclass, field
from glob import glob
from typing import Optional

import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)


@dataclass
class ArrayComparisonResult:
    almost_equal: bool
    num_diff_pixels: int
    total_pixels: int
    different_shape: bool

    @property
    def failed(self):
        return self.different_shape or not self.almost_equal

    @property
    def diff_percentage(self):
        if not self.total_pixels or self.different_shape:
            return 100.0
        return self.num_diff_pixels / self.total_pixels * 100


@dataclass
class FlatArrayComparisonResult(ArrayComparisonResult):
    # add additional information for formats that don't contain it
    shape1: tuple
    shape2: tuple
    dtype1: np.dtype
    dtype2: np.dtype


@dataclass
class VariableComparisonResult(ArrayComparisonResult):
    variable: str
    variable_missing: bool

    @property
    def failed(self):
        return super().failed or self.variable_missing


@dataclass
class FileComparisonResults:
    file1: str
    file2: str
    files_missing: bool
    unknown_file_type: bool
    sub_results: list[ArrayComparisonResult] = field(default_factory=list)

    @property
    def any_failed(self) -> bool:
        if self.files_missing or self.unknown_file_type:
            return True
        for sub_result in self.sub_results:
            if sub_result.failed:
                return True
        return False


def isclose_array(array1, array2, atol=0.0, rtol=0.0, margin_of_error=0.0, **kwargs) -> ArrayComparisonResult:
    """Compare 2 binary arrays per pixel.

    Two pixels are considered different if the absolute value of their
    difference is greater than 1. This function assumes the arrays are
    in useful data types, which may cause erroneous results. For example,
    if both arrays are unsigned integers and the different results in a
    negative value overflow will occur and the threshold will likely not
    be met.

    Args:
        array1: numpy array for comparison
        array2: numpy array for comparison
        atol: absolute tolerance (see numpy ``isclose``)
        rtol: relative tolerance (see numpy ``isclose``)
        margin_of_error: percentage of pixels that can be different and still
            be considered a passing amount.
        kwargs: Unused.


    Returns:
        1 if more than margin_of_error pixels are different, 0 otherwise.

    """
    if array1.shape != array2.shape:
        LOG.error("Data shapes were not equal: %r | %r", array1.shape, array2.shape)
        return ArrayComparisonResult(False, 0, 0, True)

    total_pixels = array1.size
    equal_pixels = np.count_nonzero(np.isclose(array1, array2, rtol=rtol, atol=atol, equal_nan=True))
    diff_pixels = total_pixels - equal_pixels
    if diff_pixels > margin_of_error / 100 * total_pixels:
        LOG.warning("%d pixels out of %d pixels are different" % (diff_pixels, total_pixels))
        return ArrayComparisonResult(False, diff_pixels, total_pixels, False)
    LOG.info("%d pixels out of %d pixels are different" % (diff_pixels, total_pixels))
    return ArrayComparisonResult(True, diff_pixels, total_pixels, False)


def plot_array(array1, array2, cmap="viridis", vmin=None, vmax=None, **kwargs):
    """Debug two arrays being different by visually comparing them."""
    import matplotlib.pyplot as plt

    LOG.info("Plotting arrays...")
    vmin = vmin if vmin else -10
    vmax = vmax if vmax else 10

    fig, (ax1, ax2) = plt.subplots(2, 2)
    array3 = array1 - array2
    q = [0, 0.25, 0.5, 0.75, 1.0]
    array3_quantiles = np.nanquantile(array3, q)
    subtitle = np.array2string(array3_quantiles, precision=8, separator=",", suppress_small=True)
    fig.suptitle(subtitle)
    img1 = ax1[0].imshow(array1, cmap=cmap, vmin=vmin, vmax=vmax)
    fig.colorbar(img1, ax=ax1[0])
    img2 = ax1[1].imshow(array2, cmap=cmap, vmin=vmin, vmax=vmax)
    fig.colorbar(img2, ax=ax1[1])

    ax2[0].set_title("Difference")
    array3[array3 == 0.0] = np.nan
    img4 = ax2[0].imshow(array3, cmap="RdBu", vmin=vmin, vmax=vmax)
    fig.colorbar(img4, ax=ax2[0])

    n_bins = 100
    nan_array3 = array3[~np.isnan(array3)]
    ax2[1].hist(nan_array3, density=True, bins=n_bins)

    plt.tight_layout()
    plt.show()


def compare_array(array1, array2, plot=False, **kwargs) -> ArrayComparisonResult:
    if plot:
        plot_array(array1, array2, **kwargs)
    return isclose_array(array1, array2, **kwargs)


def compare_binary(fn1, fn2, shape, dtype, atol=0.0, margin_of_error=0.0, **kwargs) -> list[ArrayComparisonResult]:
    if dtype is None:
        dtype = np.float32
    mmap_kwargs = {"dtype": dtype, "mode": "r"}
    if shape is not None and shape[0] is not None:
        mmap_kwargs["shape"] = shape
    array1 = np.memmap(fn1, **mmap_kwargs)
    array2 = np.memmap(fn2, **mmap_kwargs)

    array_comparison = compare_array(array1, array2, atol=atol, margin_of_error=margin_of_error, **kwargs)
    flat_array_comparison = FlatArrayComparisonResult(
        **array_comparison.__dict__, shape1=array1.shape, shape2=array2.shape, dtype1=array1.dtype, dtype2=array2.dtype
    )
    return [flat_array_comparison]


def compare_geotiff(gtiff_fn1, gtiff_fn2, atol=0.0, margin_of_error=0.0, **kwargs) -> list[ArrayComparisonResult]:
    """Compare 2 single banded geotiff files.

    .. note::

        The binary arrays will be converted to 32-bit floats before
        comparison.

    """
    array1 = _get_geotiff_array(gtiff_fn1, dtype=np.float32)
    array2 = _get_geotiff_array(gtiff_fn2, dtype=np.float32)
    arr_compare = compare_array(array1, array2, atol=atol, margin_of_error=margin_of_error, **kwargs)
    if arr_compare.failed:
        return [arr_compare]

    # array data are equal, but
    cmap1 = _get_geotiff_colormap(gtiff_fn1)
    cmap2 = _get_geotiff_colormap(gtiff_fn2)
    cmap_result = _compare_gtiff_colormaps(cmap1, cmap2, atol=atol, **kwargs)
    # only return colormap results if there was actually a colormap
    if cmap_result.total_pixels:
        return [arr_compare, cmap_result]
    return [arr_compare]


def _get_geotiff_array(gtiff_fn, dtype=None):
    import rasterio

    band_arrays = []
    with rasterio.open(gtiff_fn, "r") as gtiff_file:
        for band_idx in range(gtiff_file.count):
            arr = gtiff_file.read(band_idx + 1)
            if dtype is not None:
                arr = arr.astype(dtype)
            band_arrays.append(arr)
    return np.concatenate(band_arrays)


def _get_geotiff_colormap(gtiff_fn, band_idx=1):
    import rasterio

    with rasterio.open(gtiff_fn, "r") as gtiff_file:
        try:
            return gtiff_file.colormap(band_idx)
        except ValueError:
            return None


def _compare_gtiff_colormaps(cmap1: dict, cmap2: dict, **kwargs) -> VariableComparisonResult:
    if cmap1 is None and cmap2 is None:
        return VariableComparisonResult(True, 0, 0, False, "colormap", True)
    len1 = len(cmap1) if cmap1 is not None else 0
    len2 = len(cmap2) if cmap2 is not None else 0
    if len1 != len2:
        return VariableComparisonResult(False, abs(len2 - len1), max(len1, len2), True, "colormap", False)
    arr1 = np.array([[control_point] + list(color) for control_point, color in cmap1.items()])
    arr2 = np.array([[control_point] + list(color) for control_point, color in cmap2.items()])
    array_result = compare_array(arr1, arr2, **kwargs)
    return VariableComparisonResult(**array_result.__dict__, variable="colormap", variable_missing=False)


def compare_netcdf(
    nc1_name, nc2_name, variables, atol=0.0, margin_of_error=0.0, **kwargs
) -> list[VariableComparisonResult]:
    nc1 = xr.open_dataset(nc1_name)
    nc2 = xr.open_dataset(nc2_name)

    if variables is None:
        # TODO: Handle groups
        variables = list(nc1.variables.keys())
    results = []
    for v in variables:
        image1_var = nc1[v].data
        image2_var = nc2[v].data
        LOG.debug("Comparing data for variable '{}'".format(v))
        array_result = compare_array(image1_var, image2_var, atol=atol, margin_of_error=margin_of_error, **kwargs)
        var_result = VariableComparisonResult(**array_result.__dict__, variable=v, variable_missing=False)
        results.append(var_result)
    return results


def _get_hdf5_variables(variables, name, obj):
    import h5py

    if isinstance(obj, h5py.Group):
        return
    variables.append(name)


def compare_hdf5(h1_name, h2_name, variables, atol=0.0, margin_of_error=0.0, **kwargs) -> list[ArrayComparisonResult]:
    import h5py

    h1 = h5py.File(h1_name, "r")
    h2 = h5py.File(h2_name, "r")
    if variables is None:
        from functools import partial

        variables = []
        cb = partial(_get_hdf5_variables, variables)
        h1.visititems(cb)

    results = []
    for v in variables:
        LOG.debug("Comparing data for variable '{}'".format(v))
        image1_var = h1[v]
        if v not in h2:
            total_pixels = image1_var.size
            var_result = VariableComparisonResult(False, total_pixels, total_pixels, True, v, True)
        else:
            image2_var = h2[v]
            array_result = compare_array(
                image1_var[:], image2_var[:], atol=atol, margin_of_error=margin_of_error, **kwargs
            )
            var_result = VariableComparisonResult(**array_result.__dict__, variable=v, variable_missing=False)
        results.append(var_result)
    return results


def compare_image(im1_name, im2_name, atol=0.0, margin_of_error=0.0, **kwargs) -> list[ArrayComparisonResult]:
    img1 = _get_image_array(im1_name)
    img2 = _get_image_array(im2_name)
    return [compare_array(img1, img2, atol=atol, margin_of_error=margin_of_error, **kwargs)]


def _get_image_array(
    img_filename: str, variable: str = None, shape: Optional[tuple] = None, dtype: Optional[np.dtype] = None
) -> Optional[np.ndarray]:
    from PIL import Image

    if variable is not None:
        # NotImplementedError (geotiff colormap)
        return None

    # we may be dealing with large images that look like decompression bombs
    # let's turn off the check for the image size in PIL/Pillow
    Image.MAX_IMAGE_PIXELS = None

    img = Image.open(img_filename)
    if "P" in img.mode:
        img = img.convert("RGB" if img.mode == "P" else "RGBA")
    return np.array(img)


def _get_netcdf_array(
    input_filename: str, variable: str, shape: Optional[tuple], dtype: Optional[np.dtype]
) -> Optional[np.ndarray]:
    import xarray as xr

    ds = xr.open_dataset(input_filename)
    if variable not in ds:
        return None
    arr = ds[variable].data
    if not arr.shape:
        return None
    arr = _tranpose_for_thumbnail_if_multiband_array(arr)
    return arr


def _get_hdf5_array(
    input_filename: str, variable: str, shape: Optional[tuple], dtype: Optional[np.dtype]
) -> Optional[np.ndarray]:
    import h5py

    h = h5py.File(input_filename, "r")
    if variable is None or variable not in h:
        return None
    arr = np.array(h[variable][:])
    arr = _tranpose_for_thumbnail_if_multiband_array(arr)
    return arr


def _tranpose_for_thumbnail_if_multiband_array(arr: np.ndarray) -> np.ndarray:
    if arr.ndim == 3 and arr.shape[0] in (3, 4):
        # assume RGB/RGBA with band dimension first
        # need to transpose for PIL image RGB
        arr = arr.transpose((1, 2, 0))
    return arr


def _get_binary_array(
    input_filename: str,
    variable: str,
    shape: Optional[tuple],
    dtype: np.dtype,
) -> Optional[np.ndarray]:
    if variable is not None:
        return None
    mmap_kwargs = {"dtype": dtype, "mode": "r"}
    if shape is not None and shape[0] is not None:
        mmap_kwargs["shape"] = shape
    array1 = np.memmap(input_filename, **mmap_kwargs)
    return array1


type_name_to_compare_func = {
    "binary": compare_binary,
    "gtiff": compare_geotiff,
    "geotiff": compare_geotiff,
    "netcdf": compare_netcdf,
    "hdf5": compare_hdf5,
    "png": compare_image,
}

file_ext_to_compare_func = {
    ".dat": compare_binary,
    ".tif": compare_geotiff,
    ".tiff": compare_geotiff,
    ".nc": compare_netcdf,
    ".h5": compare_hdf5,
    ".png": compare_image,
    ".jpg": compare_image,
    ".jpeg": compare_image,
}

file_ext_to_array_func = {
    ".dat": _get_binary_array,
    ".nc": _get_netcdf_array,
    ".h5": _get_hdf5_array,
    ".tif": _get_image_array,
    ".png": _get_image_array,
    ".jpg": _get_image_array,
    ".jpeg": _get_image_array,
}


def _file_type(str_val):
    str_val = str_val.lower()
    if str_val in type_name_to_compare_func:
        return type_name_to_compare_func[str_val]

    print("ERROR: Unknown file type '%s'" % (str_val,))
    print("Possible file types: \n\t%s" % ("\n\t".join(type_name_to_compare_func.keys())))
    raise ValueError("Unknown file type '%s'" % (str_val,))


class CompareHelper:
    """Wrapper around various comparison operations."""

    def __init__(self, atol: float = 0.0, rtol: float = 0.0, margin_of_error: float = 0.0, create_plot: bool = False):
        self.atol = atol
        self.rtol = rtol
        self.margin_of_error = margin_of_error
        self.create_plot = create_plot

    def compare_files(self, file1, file2, file_type=None, **kwargs):
        if file_type is None:
            # guess based on file extension
            ext = os.path.splitext(file1)[-1]
            file_type = file_ext_to_compare_func.get(ext)
        if file_type is None:
            LOG.error(f"Could not determine how to compare file type (extension not recognized): {file1}.")
            return FileComparisonResults(file1, file2, False, True)
        LOG.info(f"Comparing {file2!r} to known valid file {file1!r}.")
        comparison_results = file_type(
            file1,
            file2,
            atol=self.atol,
            rtol=self.rtol,
            margin_of_error=self.margin_of_error,
            plot=self.create_plot,
            **kwargs,
        )
        return FileComparisonResults(file1, file2, False, False, comparison_results)

    def compare_dirs(self, dir1, dir2, **kwargs) -> list[FileComparisonResults]:
        results = []
        for expected_path in sorted(glob(os.path.join(dir1, "*"))):
            if expected_path.endswith(".log"):
                continue
            test_path = os.path.join(dir2, os.path.basename(expected_path))
            if not os.path.isfile(test_path):
                LOG.error(f"File from first directory is not present in second directory: {test_path}")
                results.append(FileComparisonResults(expected_path, test_path, True, False))
                continue

            file_comparison_results = self.compare(expected_path, test_path, **kwargs)
            results.extend(file_comparison_results)
        return results

    def compare(self, input1, input2, **kwargs) -> list[FileComparisonResults]:
        if os.path.isdir(input1) and os.path.isdir(input2):
            return self.compare_dirs(input1, input2, **kwargs)
        elif os.path.isfile(input1) and os.path.isfile(input2):
            return [self.compare_files(input1, input2, **kwargs)]
        elif not os.path.exists(input1):
            LOG.error("Could not find input directory or file {}".format(input1))
            return [FileComparisonResults(input1, input2, True, False)]
        elif not os.path.exists(input2):
            LOG.error("Could not find input directory or file {}".format(input2))
            return [FileComparisonResults(input1, input2, True, False)]
        else:
            LOG.error("Inputs are not both files or both directories.")
            return [FileComparisonResults(input1, input2, True, False)]


def num_failed_files(file_comparison_results: list[FileComparisonResults]) -> int:
    return sum(int(fc.any_failed) for fc in file_comparison_results)


HTML_TEMPLATE = """
<html lang="en">
<head>
<title>{title}</title>
</head>

<h1>{title}</h1>

<table>
<tr>
    <th>Filename</th>
    <th>Status</th>
    <th>Variable</th>
    <th>Expected</th>
    <th>Actual</th>
    <th>Diff. Pixels (%)</th>
    <th>Notes</th>
</tr>
{rows}
</table>

<body>
</body>
</html>
"""


def _generate_html_summary(output_filename, file_comparison_results):
    filename = os.path.basename(output_filename)
    row_html = "\n\t" + "\n\t".join(_generate_table_rows(output_filename, file_comparison_results))
    LOG.info(f"Creating HTML file {output_filename}")
    dst_dir = os.path.dirname(output_filename)
    if dst_dir:
        os.makedirs(dst_dir, exist_ok=True)
    with open(output_filename, "w") as html_file:
        html_text = HTML_TEMPLATE.format(
            title=filename,
            rows=row_html,
        )
        html_file.write(html_text)


def _generate_table_rows(
    output_filename: str,
    file_comparison_results: list[FileComparisonResults],
) -> list[str]:
    img_dst_dir = os.path.join(os.path.dirname(output_filename), "_images")
    os.makedirs(img_dst_dir, exist_ok=True)
    row_infos = []
    for fc in file_comparison_results:
        exp_filename = os.path.basename(fc.file1)
        if fc.files_missing or fc.unknown_file_type:
            row_info = ROW_TEMPLATE.format(
                filename=exp_filename,
                status="FAILED",
                variable="N/A",
                expected_img="N/A",
                actual_img="N/A",
                diff_percent=100.0,
                notes="Missing file" if fc.files_missing else "Unknown file type",
            )
            row_infos.append(row_info)
            continue

        for sub_result in fc.sub_results:
            row_info = _generate_subresult_table_row(img_dst_dir, fc, sub_result)
            row_infos.append(row_info)
    return row_infos


ROW_TEMPLATE = """
<tr>
    <td>{filename}</td>
    <td>{status}</td>
    <td>{variable}</td>
    <td>{expected_img}</td>
    <td>{actual_img}</td>
    <td>{diff_percent:0.02f}%</td>
    <td>{notes}</td>
</tr>
"""


def _generate_subresult_table_row(
    img_dst_dir: str,
    file_comparison_result: FileComparisonResults,
    sub_result: ArrayComparisonResult,
) -> str:
    status = "FAILED" if sub_result.failed else "PASSED"
    notes = ""
    diff_percent = sub_result.diff_percentage
    if sub_result.different_shape:
        notes = "Different array shapes"
    elif sub_result.failed:
        notes = "Too many differing pixels"

    variable = getattr(sub_result, "variable", None)
    exp_tn_html = _generate_thumbnail_html(
        file_comparison_result.file1,
        variable,
        img_dst_dir,
        "expected",
        getattr(sub_result, "shape1", None),
        getattr(sub_result, "dtype1", None),
    )
    act_tn_html = _generate_thumbnail_html(
        file_comparison_result.file2,
        variable,
        img_dst_dir,
        "actual",
        getattr(sub_result, "shape2", None),
        getattr(sub_result, "dtype2", None),
    )
    exp_filename = os.path.basename(file_comparison_result.file1)
    row_info = ROW_TEMPLATE.format(
        filename=exp_filename,
        status=status,
        variable=variable or "N/A",
        actual_img=act_tn_html,
        expected_img=exp_tn_html,
        diff_percent=diff_percent,
        notes=notes,
    )
    return row_info


def _generate_thumbnail_html(
    data_pathname: str,
    variable: Optional[str],
    img_dst_dir: str,
    tn_suffix: str,
    shape: Optional[tuple],
    dtype: Optional[np.dtype],
) -> str:
    data_arr = _get_thumbnail_array(data_pathname, variable, shape, dtype)
    if data_arr is None:
        return "N/A"
    data_filename = os.path.basename(data_pathname)
    file_ext = os.path.splitext(data_filename)[1]
    var_name = variable.replace("/", "-") if variable is not None else "None"
    exp_tn_fn = data_filename.replace(file_ext, f".{var_name}.{tn_suffix}.png")
    try:
        _generate_thumbnail(data_arr, os.path.join(img_dst_dir, exp_tn_fn), max_width=512)
    except (RuntimeError, ValueError):
        return "Failed to generate thumbnail"
    exp_tn_html = IMG_ENTRY_TMPL.format("_images/" + exp_tn_fn)
    return exp_tn_html


def _get_thumbnail_array(
    input_data_path: str, variable: Optional[str], shape: Optional[tuple], dtype: Optional[np.dtype]
) -> Optional[np.ndarray]:
    input_ext = os.path.splitext(input_data_path)[1]
    if input_ext not in file_ext_to_array_func:
        return None
    input_arr = file_ext_to_array_func[input_ext](input_data_path, variable, shape, dtype)
    return input_arr


IMG_ENTRY_TMPL = '<img src="{}"></img>'


def _generate_thumbnail(input_arr, output_thumbnail_path, max_width=512) -> bool:
    if input_arr.ndim == 1:
        return _generate_matplotlib_1d_thumbnail(input_arr, output_thumbnail_path, max_width)
    if input_arr.dtype == np.uint8:
        return _generate_pillow_thumbnail(input_arr, output_thumbnail_path, max_width)
    return _generate_matplotlib_thumbnail(input_arr, output_thumbnail_path, max_width)


def _generate_matplotlib_1d_thumbnail(input_arr, output_thumbnail_path, max_width) -> bool:
    import matplotlib.pyplot as plt

    figsize = _get_mpl_figsize(input_arr.shape, max_width)
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(input_arr, bins=10)
    fig.savefig(output_thumbnail_path)
    plt.close(fig)
    return True


def _generate_matplotlib_thumbnail(input_arr, output_thumbnail_path, max_width=512) -> bool:
    import matplotlib.pyplot as plt

    figsize = _get_mpl_figsize(input_arr.shape, max_width)
    fig, ax = plt.subplots(figsize=figsize)
    img = ax.imshow(input_arr)
    fig.colorbar(img, ax=ax)
    fig.savefig(output_thumbnail_path)
    plt.close(fig)
    return True


def _get_mpl_figsize(input_shape, max_width) -> tuple[int, int]:
    # dpi 100 => 51
    fig_width = max_width / 100.0
    if len(input_shape) == 1:
        fig_height = fig_width
    else:
        fig_height = input_shape[0] * (fig_width / input_shape[1])
    return fig_width, fig_height


def _generate_pillow_thumbnail(input_arr, output_thumbnail_path, max_width=512) -> bool:
    from PIL import Image

    # we may be dealing with large images that look like decompression bombs
    # let's turn off the check for the image size in PIL/Pillow
    Image.MAX_IMAGE_PIXELS = None

    full_img = Image.fromarray(input_arr)
    full_size = full_img.size
    max_width = min(full_size[0], max_width)
    width_ratio = full_size[0] // max_width
    new_size = (max_width, full_size[1] // width_ratio)
    scaled_img = full_img.resize(new_size)
    scaled_img.save(output_thumbnail_path, format="PNG")
    return True


def main(argv=sys.argv[1:]):
    from argparse import ArgumentParser

    from polar2grid.core.dtype import str_to_dtype

    parser = ArgumentParser(description="Compare two files per pixel")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG (default ERROR)",
    )
    parser.add_argument(
        "--atol",
        type=float,
        default=0.0,
        help="specify threshold for comparison differences (see numpy.isclose 'atol' parameter)",
    )
    parser.add_argument(
        "--rtol",
        type=float,
        default=0.0,
        help="specify relative tolerance for comparison (see numpy.isclose 'rtol' parameter)",
    )
    parser.add_argument(
        "--shape",
        dest="shape",
        type=int,
        nargs=2,
        default=(None, None),
        help="'rows cols' for binary file comparison only",
    )
    parser.add_argument("--dtype", type=str_to_dtype, help="Data type for binary file comparison only")
    parser.add_argument(
        "--variables",
        nargs="+",
        help="NetCDF/HDF5 variables to read and compare. If not provided all variables will be compared.",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        help="Show a plot of the two arrays instead of checking equality. Used for debugging.",
    )
    parser.add_argument(
        "--html",
        nargs="?",
        default=False,
        help="Generate an HTML page summarizing comparison status and save it to this filename. "
        "If specified with no argument then defaults to 'comparison_summary.html'. All additional "
        "files (images, CSS, etc) will be placed in the same directory.",
    )
    parser.add_argument("--margin-of-error", type=float, default=0.0, help="percent of total pixels that can be wrong")
    parser.add_argument(
        "file_type",
        type=_file_type,
        nargs="?",
        help="Type of files being compare. If not provided it will be determined based on file extension.",
    )
    parser.add_argument("input1", help="First filename or directory to compare. This is typically the expected output.")
    parser.add_argument("input2", help="Second filename or directory to compare. This is typicall the actual output.")
    args = parser.parse_args(argv)

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)])
    compare_kwargs = {
        "shape": tuple(args.shape),
        "dtype": args.dtype,
        "variables": args.variables,
        "file_type": args.file_type,
    }

    comparer = CompareHelper(
        atol=args.atol, rtol=args.rtol, margin_of_error=args.margin_of_error, create_plot=args.plot
    )
    file_comparison_results = comparer.compare(args.input1, args.input2, **compare_kwargs)
    num_files = num_failed_files(file_comparison_results)

    if args.html is None:
        args.html = "comparison_summary.html"
    if args.html:
        _generate_html_summary(args.html, file_comparison_results)

    if num_files == 0:
        print("All files passed")
        print("SUCCESS")
    else:
        print(f"ERROR: {num_files} files were found to be unequal")

    # 0 if successful, otherwise number of failed files
    return num_files


if __name__ == "__main__":
    sys.exit(main())
